@@ -21,7 +21,6 @@
  * Copyright (C) 2013-2014 Ping Identity Corporation
  * All rights reserved.
  *
- * The contents of this file are the property of Ping Identity Corporation.
  * For further information please contact:
  *
  *      Ping Identity Corporation
@@ -93,18 +92,21 @@ int oidc_cache_file_post_config(server_rec *s) {
 /*
  * return the cache file name for a specified key
  */
-static const char *oidc_cache_file_name(request_rec *r, const char *key) {
-	return apr_psprintf(r->pool, "%s%s", OIDC_CACHE_FILE_PREFIX, key);
+static const char *oidc_cache_file_name(request_rec *r, const char *section,
+		const char *key) {
+	return apr_psprintf(r->pool, "%s%s-%s", OIDC_CACHE_FILE_PREFIX, section,
+			key);
 }
 
 /*
  * return the fully qualified path name to a cache file for a specified key
  */
-static const char *oidc_cache_file_path(request_rec *r, const char *key) {
+static const char *oidc_cache_file_path(request_rec *r, const char *section,
+		const char *key) {
 	oidc_cfg *cfg = ap_get_module_config(r->server->module_config,
 			&auth_openidc_module);
 	return apr_psprintf(r->pool, "%s/%s", cfg->cache_file_dir,
-			oidc_cache_file_name(r, key));
+			oidc_cache_file_name(r, section, key));
 }
 
 /*
@@ -122,15 +124,14 @@ static apr_status_t oidc_cache_file_read(request_rec *r, const char *path,
 
 	/* test for system errors */
 	if (rc != APR_SUCCESS) {
-		ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, r,
-				"oidc_cache_file_read: could not read from: %s (%s)", path,
+		oidc_error(r, "could not read from: %s (%s)", path,
 				apr_strerror(rc, s_err, sizeof(s_err)));
 	}
 
 	/* ensure that we've got the requested number of bytes */
 	if (bytes_read != len) {
-		ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, r,
-				"oidc_cache_file_read: could not read enough bytes from: \"%s\", bytes_read (%" APR_SIZE_T_FMT ") != len (%" APR_SIZE_T_FMT ")",
+		oidc_error(r,
+				"could not read enough bytes from: \"%s\", bytes_read (%" APR_SIZE_T_FMT ") != len (%" APR_SIZE_T_FMT ")",
 				path, bytes_read, len);
 		rc = APR_EGENERAL;
 	}
@@ -153,16 +154,15 @@ static apr_status_t oidc_cache_file_write(request_rec *r, const char *path,
 
 	/* check for a system error */
 	if (rc != APR_SUCCESS) {
-		ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, r,
-				"oidc_cache_file_write: could not write to: \"%s\" (%s)", path,
+		oidc_error(r, "could not write to: \"%s\" (%s)", path,
 				apr_strerror(rc, s_err, sizeof(s_err)));
 		return rc;
 	}
 
 	/* check that all bytes from the header were written */
 	if (bytes_written != len) {
-		ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, r,
-				"oidc_cache_file_write: could not write enough bytes to: \"%s\", bytes_written (%" APR_SIZE_T_FMT ") != len (%" APR_SIZE_T_FMT ")",
+		oidc_error(r,
+				"could not write enough bytes to: \"%s\", bytes_written (%" APR_SIZE_T_FMT ") != len (%" APR_SIZE_T_FMT ")",
 				path, bytes_written, len);
 		return APR_EGENERAL;
 	}
@@ -173,20 +173,19 @@ static apr_status_t oidc_cache_file_write(request_rec *r, const char *path,
 /*
  * get a value for the specified key from the cache
  */
-static apr_byte_t oidc_cache_file_get(request_rec *r, const char *key,
-		const char **value) {
+static apr_byte_t oidc_cache_file_get(request_rec *r, const char *section,
+		const char *key, const char **value) {
 	apr_file_t *fd = NULL;
 	apr_status_t rc = APR_SUCCESS;
 	char s_err[128];
 
 	/* get the fully qualified path to the cache file based on the key name */
-	const char *path = oidc_cache_file_path(r, key);
+	const char *path = oidc_cache_file_path(r, section, key);
 
 	/* open the cache file if it exists, otherwise we just have a "regular" cache miss */
 	if (apr_file_open(&fd, path, APR_FOPEN_READ | APR_FOPEN_BUFFERED,
 	APR_OS_DEFAULT, r->pool) != APR_SUCCESS) {
-		ap_log_rerror(APLOG_MARK, OIDC_DEBUG, 0, r,
-				"oidc_cache_file_get: cache miss for key \"%s\"", key);
+		oidc_debug(r, "cache miss for key \"%s\"", key);
 		return TRUE;
 	}
 
@@ -211,15 +210,13 @@ static apr_byte_t oidc_cache_file_get(request_rec *r, const char *key,
 		apr_file_close(fd);
 
 		/* log this event */
-		ap_log_rerror(APLOG_MARK, OIDC_DEBUG, 0, r,
-				"oidc_cache_file_get: cache entry \"%s\" expired, removing file \"%s\"",
-				key, path);
+		oidc_debug(r, "cache entry \"%s\" expired, removing file \"%s\"", key,
+				path);
 
 		/* and kill it */
 		if ((rc = apr_file_remove(path, r->pool)) != APR_SUCCESS) {
-			ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, r,
-					"oidc_cache_file_get: could not delete cache file \"%s\" (%s)",
-					path, apr_strerror(rc, s_err, sizeof(s_err)));
+			oidc_error(r, "could not delete cache file \"%s\" (%s)", path,
+					apr_strerror(rc, s_err, sizeof(s_err)));
 		}
 
 		/* nothing strange happened really */
@@ -234,9 +231,7 @@ static apr_byte_t oidc_cache_file_get(request_rec *r, const char *key,
 
 	/* barf on failure */
 	if (rc != APR_SUCCESS) {
-		ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, r,
-				"oidc_cache_file_get: could not read cache value from \"%s\"",
-				path);
+		oidc_error(r, "could not read cache value from \"%s\"", path);
 		goto error_close;
 	}
 
@@ -245,8 +240,8 @@ static apr_byte_t oidc_cache_file_get(request_rec *r, const char *key,
 	apr_file_close(fd);
 
 	/* log a successful cache hit */
-	ap_log_rerror(APLOG_MARK, OIDC_DEBUG, 0, r,
-			"oidc_cache_file_get: cache hit for key \"%s\" (%" APR_SIZE_T_FMT " bytes, expiring in: %" APR_TIME_T_FMT ")",
+	oidc_debug(r,
+			"cache hit for key \"%s\" (%" APR_SIZE_T_FMT " bytes, expiring in: %" APR_TIME_T_FMT ")",
 			key, info.len, apr_time_sec(info.expire - apr_time_now()));
 
 	return TRUE;
@@ -256,8 +251,7 @@ error_close:
 	apr_file_unlock(fd);
 	apr_file_close(fd);
 
-	ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, r,
-			"oidc_cache_file_get: return error status %d (%s)", rc,
+	oidc_error(r, "return error status %d (%s)", rc,
 			apr_strerror(rc, s_err, sizeof(s_err)));
 
 	return FALSE;
@@ -282,8 +276,8 @@ static apr_status_t oidc_cache_file_clean(request_rec *r) {
 			&auth_openidc_module);
 
 	/* get the path to the metadata file that holds "last cleaned" metadata info */
-	const char *metadata_path = oidc_cache_file_path(r,
-	OIDC_CACHE_FILE_LAST_CLEANED);
+	const char *metadata_path = oidc_cache_file_path(r, "cache-file",
+			OIDC_CACHE_FILE_LAST_CLEANED);
 
 	/* open the metadata file if it exists */
 	if ((rc = apr_stat(&fi, metadata_path, APR_FINFO_MTIME, r->pool))
@@ -291,11 +285,10 @@ static apr_status_t oidc_cache_file_clean(request_rec *r) {
 
 		/* really only clean once per so much time, check that we haven not recently run */
 		if (apr_time_now() < fi.mtime + apr_time_from_sec(cfg->cache_file_clean_interval)) {
-			ap_log_rerror(APLOG_MARK, OIDC_DEBUG, 0, r,
-					"oidc_cache_clean: last cleanup call was less than %d seconds ago (next one as early as in %" APR_TIME_T_FMT " secs)",
+			oidc_debug(r,
+					"last cleanup call was less than %d seconds ago (next one as early as in %" APR_TIME_T_FMT " secs)",
 					cfg->cache_file_clean_interval,
-					apr_time_sec(
-							fi.mtime + apr_time_from_sec(cfg->cache_file_clean_interval) - apr_time_now()));
+					apr_time_sec( fi.mtime + apr_time_from_sec(cfg->cache_file_clean_interval) - apr_time_now()));
 			return APR_SUCCESS;
 		}
 
@@ -308,24 +301,21 @@ static apr_status_t oidc_cache_file_clean(request_rec *r) {
 		if ((rc = apr_file_open(&fd, metadata_path,
 				(APR_FOPEN_WRITE | APR_FOPEN_CREATE), APR_OS_DEFAULT, r->pool))
 				!= APR_SUCCESS) {
-			ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, r,
-					"oidc_cache_file_clean: error creating cache timestamp file '%s' (%s)",
+			oidc_error(r, "error creating cache timestamp file '%s' (%s)",
 					metadata_path, apr_strerror(rc, s_err, sizeof(s_err)));
 			return rc;
 		}
 
 		/* and cleanup... */
 		if ((rc = apr_file_close(fd)) != APR_SUCCESS) {
-			ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, r,
-					"oidc_cache_file_clean: error closing cache timestamp file '%s' (%s)",
+			oidc_error(r, "error closing cache timestamp file '%s' (%s)",
 					metadata_path, apr_strerror(rc, s_err, sizeof(s_err)));
 		}
 	}
 
 	/* time to clean, open the cache directory */
 	if ((rc = apr_dir_open(&dir, cfg->cache_file_dir, r->pool)) != APR_SUCCESS) {
-		ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, r,
-				"oidc_cache_file_clean: error opening cache directory '%s' for cleaning (%s)",
+		oidc_error(r, "error opening cache directory '%s' for cleaning (%s)",
 				cfg->cache_file_dir, apr_strerror(rc, s_err, sizeof(s_err)));
 		return rc;
 	}
@@ -341,8 +331,9 @@ static apr_status_t oidc_cache_file_clean(request_rec *r) {
 			/* skip non-cache entries, cq. the ".", ".." and the metadata file */
 			if ((fi.name[0] == '.')
 					|| (strstr(fi.name, OIDC_CACHE_FILE_PREFIX) != fi.name)
-					|| ((apr_strnatcmp(fi.name, oidc_cache_file_name(r,
-					OIDC_CACHE_FILE_LAST_CLEANED)) == 0)))
+					|| ((apr_strnatcmp(fi.name,
+							oidc_cache_file_name(r, "cache-file",
+									OIDC_CACHE_FILE_LAST_CLEANED)) == 0)))
 				continue;
 
 			/* get the fully qualified path to the cache file and open it */
@@ -350,9 +341,8 @@ static apr_status_t oidc_cache_file_clean(request_rec *r) {
 					cfg->cache_file_dir, fi.name);
 			if ((rc = apr_file_open(&fd, path, APR_FOPEN_READ, APR_OS_DEFAULT,
 					r->pool)) != APR_SUCCESS) {
-				ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, r,
-						"oidc_cache_file_clean: unable to open cache entry \"%s\" (%s)",
-						path, apr_strerror(rc, s_err, sizeof(s_err)));
+				oidc_error(r, "unable to open cache entry \"%s\" (%s)", path,
+						apr_strerror(rc, s_err, sizeof(s_err)));
 				continue;
 			}
 
@@ -368,15 +358,14 @@ static apr_status_t oidc_cache_file_clean(request_rec *r) {
 					continue;
 
 				/* the cache entry expired, we're going to remove it so log that event */
-				ap_log_rerror(APLOG_MARK, OIDC_DEBUG, 0, r,
-						"oidc_cache_file_clean: cache entry (%s) expired, removing file \"%s\")",
+				oidc_debug(r, "cache entry (%s) expired, removing file \"%s\")",
 						fi.name, path);
 
 			} else {
 
 				/* file open returned an error, log that */
-				ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, r,
-						"oidc_cache_file_clean: cache entry (%s) corrupted (%s), removing file \"%s\"",
+				oidc_error(r,
+						"cache entry (%s) corrupted (%s), removing file \"%s\"",
 						fi.name, apr_strerror(rc, s_err, sizeof(s_err)), path);
 
 			}
@@ -385,9 +374,8 @@ static apr_status_t oidc_cache_file_clean(request_rec *r) {
 			if ((rc = apr_file_remove(path, r->pool)) != APR_SUCCESS) {
 
 				/* hrm, this will most probably happen again on the next run... */
-				ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, r,
-						"oidc_cache_file_clean: could not delete cache file \"%s\" (%s)",
-						path, apr_strerror(rc, s_err, sizeof(s_err)));
+				oidc_error(r, "could not delete cache file \"%s\" (%s)", path,
+						apr_strerror(rc, s_err, sizeof(s_err)));
 			}
 
 		}
@@ -402,14 +390,14 @@ static apr_status_t oidc_cache_file_clean(request_rec *r) {
 /*
  * write a value for the specified key to the cache
  */
-static apr_byte_t oidc_cache_file_set(request_rec *r, const char *key,
-		const char *value, apr_time_t expiry) {
+static apr_byte_t oidc_cache_file_set(request_rec *r, const char *section,
+		const char *key, const char *value, apr_time_t expiry) {
 	apr_file_t *fd = NULL;
 	apr_status_t rc = APR_SUCCESS;
 	char s_err[128];
 
 	/* get the fully qualified path to the cache file based on the key name */
-	const char *path = oidc_cache_file_path(r, key);
+	const char *path = oidc_cache_file_path(r, section, key);
 
 	/* only on writes (not on reads) we clean the cache first (if not done recently) */
 	oidc_cache_file_clean(r);
@@ -417,9 +405,8 @@ static apr_byte_t oidc_cache_file_set(request_rec *r, const char *key,
 	/* just remove cache file if value is NULL */
 	if (value == NULL) {
 		if ((rc = apr_file_remove(path, r->pool)) != APR_SUCCESS) {
-			ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, r,
-					"oidc_cache_file_set: could not delete cache file \"%s\" (%s)",
-					path, apr_strerror(rc, s_err, sizeof(s_err)));
+			oidc_error(r, "could not delete cache file \"%s\" (%s)", path,
+					apr_strerror(rc, s_err, sizeof(s_err)));
 		}
 		return TRUE;
 	}
@@ -427,9 +414,8 @@ static apr_byte_t oidc_cache_file_set(request_rec *r, const char *key,
 	/* try to open the cache file for writing, creating it if it does not exist */
 	if ((rc = apr_file_open(&fd, path, (APR_FOPEN_WRITE | APR_FOPEN_CREATE),
 	APR_OS_DEFAULT, r->pool)) != APR_SUCCESS) {
-		ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, r,
-				"oidc_cache_file_set: cache file \"%s\" could not be opened (%s)",
-				path, apr_strerror(rc, s_err, sizeof(s_err)));
+		oidc_error(r, "cache file \"%s\" could not be opened (%s)", path,
+				apr_strerror(rc, s_err, sizeof(s_err)));
 		return FALSE;
 	}
 
@@ -458,8 +444,8 @@ static apr_byte_t oidc_cache_file_set(request_rec *r, const char *key,
 	apr_file_close(fd);
 
 	/* log our success */
-	ap_log_rerror(APLOG_MARK, OIDC_DEBUG, 0, r,
-			"oidc_cache_file_set: set entry for key \"%s\" (%" APR_SIZE_T_FMT " bytes, expires in: %" APR_TIME_T_FMT ")",
+	oidc_debug(r,
+			"set entry for key \"%s\" (%" APR_SIZE_T_FMT " bytes, expires in: %" APR_TIME_T_FMT ")",
 			key, info.len, apr_time_sec(expiry - apr_time_now()));
 
 	return TRUE;
