@@ -21,7 +21,6 @@
  * Copyright (C) 2013-2014 Ping Identity Corporation
  * All rights reserved.
  *
- * The contents of this file are the property of Ping Identity Corporation.
  * For further information please contact:
  *
  *      Ping Identity Corporation
@@ -73,7 +72,8 @@ typedef struct oidc_cache_cfg_memcache_t {
 
 /* create the cache context */
 static void *oidc_cache_memcache_cfg_create(apr_pool_t *pool) {
-	oidc_cache_cfg_memcache_t *context = apr_pcalloc(pool, sizeof(oidc_cache_cfg_memcache_t));
+	oidc_cache_cfg_memcache_t *context = apr_pcalloc(pool,
+			sizeof(oidc_cache_cfg_memcache_t));
 	context->cache_memcache = NULL;
 	return context;
 }
@@ -82,9 +82,14 @@ static void *oidc_cache_memcache_cfg_create(apr_pool_t *pool) {
  * initialize the memcache struct to a number of memcache servers
  */
 static int oidc_cache_memcache_post_config(server_rec *s) {
-	oidc_cfg *cfg = (oidc_cfg *) ap_get_module_config(
-			s->module_config, &auth_openidc_module);
-	oidc_cache_cfg_memcache_t *context = (oidc_cache_cfg_memcache_t *)cfg->cache_cfg;
+	oidc_cfg *cfg = (oidc_cfg *) ap_get_module_config(s->module_config,
+			&auth_openidc_module);
+
+	if (cfg->cache_cfg != NULL)
+		return APR_SUCCESS;
+	oidc_cache_cfg_memcache_t *context = oidc_cache_memcache_cfg_create(
+			s->process->pool);
+	cfg->cache_cfg = context;
 
 	apr_status_t rv = APR_SUCCESS;
 	int nservers = 0;
@@ -93,8 +98,8 @@ static int oidc_cache_memcache_post_config(server_rec *s) {
 	apr_pool_t *p = s->process->pool;
 
 	if (cfg->cache_memcache_servers == NULL) {
-		ap_log_error(APLOG_MARK, APLOG_ERR, 0, s,
-				"oidc_cache_memcache_post_config: cache type is set to \"memcache\", but no valid OIDCMemCacheServers setting was found");
+		oidc_serror(s,
+				"cache type is set to \"memcache\", but no valid OIDCMemCacheServers setting was found");
 		return HTTP_INTERNAL_SERVER_ERROR;
 	}
 
@@ -109,8 +114,7 @@ static int oidc_cache_memcache_post_config(server_rec *s) {
 	/* allocated space for the number of servers */
 	rv = apr_memcache_create(p, nservers, 0, &context->cache_memcache);
 	if (rv != APR_SUCCESS) {
-		ap_log_error(APLOG_MARK, APLOG_ERR, rv, s,
-				"oidc_cache_memcache_init: failed to create memcache object of '%d' size",
+		oidc_serror(s, "failed to create memcache object of '%d' size",
 				nservers);
 		return HTTP_INTERNAL_SERVER_ERROR;
 	}
@@ -127,16 +131,14 @@ static int oidc_cache_memcache_post_config(server_rec *s) {
 		/* parse out host and port */
 		rv = apr_parse_addr_port(&host_str, &scope_id, &port, split, p);
 		if (rv != APR_SUCCESS) {
-			ap_log_error(APLOG_MARK, APLOG_ERR, rv, s,
-					"oidc_cache_memcache_init: failed to parse cache server: '%s'",
-					split);
+			oidc_serror(s, "failed to parse cache server: '%s'", split);
 			return HTTP_INTERNAL_SERVER_ERROR;
 		}
 
 		if (host_str == NULL) {
-			ap_log_error(APLOG_MARK, APLOG_ERR, rv, s,
-					"oidc_cache_memcache_init: failed to parse cache server, "
-							"no hostname specified: '%s'", split);
+			oidc_serror(s,
+					"failed to parse cache server, no hostname specified: '%s'",
+					split);
 			return HTTP_INTERNAL_SERVER_ERROR;
 		}
 
@@ -147,18 +149,15 @@ static int oidc_cache_memcache_post_config(server_rec *s) {
 		// TODO: tune this
 		rv = apr_memcache_server_create(p, host_str, port, 0, 1, 1, 60, &st);
 		if (rv != APR_SUCCESS) {
-			ap_log_error(APLOG_MARK, APLOG_ERR, rv, s,
-					"oidc_cache_memcache_init: failed to create cache server: %s:%d",
-					host_str, port);
+			oidc_serror(s, "failed to create cache server: %s:%d", host_str,
+					port);
 			return HTTP_INTERNAL_SERVER_ERROR;
 		}
 
 		/* add the memcache server struct to the list */
 		rv = apr_memcache_add_server(context->cache_memcache, st);
 		if (rv != APR_SUCCESS) {
-			ap_log_error(APLOG_MARK, APLOG_ERR, rv, s,
-					"oidc_cache_memcache_init: failed to add cache server: %s:%d",
-					host_str, port);
+			oidc_serror(s, "failed to add cache server: %s:%d", host_str, port);
 			return HTTP_INTERNAL_SERVER_ERROR;
 		}
 
@@ -169,36 +168,45 @@ static int oidc_cache_memcache_post_config(server_rec *s) {
 	return OK;
 }
 
+/*
+ * assemble single key name based on section/key input
+ */
+static char *oidc_cache_memcache_get_key(apr_pool_t *pool, const char *section,
+		const char *key) {
+	return apr_psprintf(pool, "%s:%s", section, key);
+}
+
 /*
  * get a name/value pair from memcache
  */
-static apr_byte_t oidc_cache_memcache_get(request_rec *r, const char *key,
-		const char **value) {
+static apr_byte_t oidc_cache_memcache_get(request_rec *r, const char *section,
+		const char *key, const char **value) {
 
-	ap_log_rerror(APLOG_MARK, OIDC_DEBUG, 0, r,
-			"oidc_cache_memcache_get: entering \"%s\"", key);
+	oidc_debug(r, "enter, section=\"%s\", key=\"%s\"", section, key);
 
 	oidc_cfg *cfg = ap_get_module_config(r->server->module_config,
 			&auth_openidc_module);
-	oidc_cache_cfg_memcache_t *context = (oidc_cache_cfg_memcache_t *)cfg->cache_cfg;
+	oidc_cache_cfg_memcache_t *context =
+			(oidc_cache_cfg_memcache_t *) cfg->cache_cfg;
 
 	apr_size_t len = 0;
 
 	/* get it */
-	apr_status_t rv = apr_memcache_getp(context->cache_memcache, r->pool, key,
-			(char **)value, &len, NULL);
+	apr_status_t rv = apr_memcache_getp(context->cache_memcache, r->pool,
+			oidc_cache_memcache_get_key(r->pool, section, key), (char **) value,
+			&len, NULL);
 
 	// TODO: error strings ?
 	if (rv != APR_SUCCESS) {
-		ap_log_rerror(APLOG_MARK, APLOG_ERR, rv, r,
-				"oidc_cache_memcache_get: apr_memcache_getp returned an error");
+		oidc_error(r, "apr_memcache_getp returned an error");
 		return FALSE;
 	}
 
 	/* do sanity checking on the string value */
-	if ( (*value) && (strlen(*value) != len) ) {
-		ap_log_rerror(APLOG_MARK, APLOG_ERR, rv, r,
-				"oidc_cache_memcache_get: apr_memcache_getp returned less bytes than expected: strlen(value) [%zu] != len [%" APR_SIZE_T_FMT "]", strlen(*value), len);
+	if ((*value) && (strlen(*value) != len)) {
+		oidc_error(r,
+				"apr_memcache_getp returned less bytes than expected: strlen(value) [%zu] != len [%" APR_SIZE_T_FMT "]",
+				strlen(*value), len);
 		return FALSE;
 	}
 
@@ -208,27 +216,27 @@ static apr_byte_t oidc_cache_memcache_get(request_rec *r, const char *key,
 /*
  * store a name/value pair in memcache
  */
-static apr_byte_t oidc_cache_memcache_set(request_rec *r, const char *key,
-		const char *value, apr_time_t expiry) {
+static apr_byte_t oidc_cache_memcache_set(request_rec *r, const char *section,
+		const char *key, const char *value, apr_time_t expiry) {
 
-	ap_log_rerror(APLOG_MARK, OIDC_DEBUG, 0, r,
-			"oidc_cache_memcache_set: entering \"%s\"", key);
+	oidc_debug(r, "enter, section=\"%s\", key=\"%s\"", section, key);
 
 	oidc_cfg *cfg = ap_get_module_config(r->server->module_config,
 			&auth_openidc_module);
-	oidc_cache_cfg_memcache_t *context = (oidc_cache_cfg_memcache_t *)cfg->cache_cfg;
+	oidc_cache_cfg_memcache_t *context =
+			(oidc_cache_cfg_memcache_t *) cfg->cache_cfg;
 
 	apr_status_t rv = APR_SUCCESS;
 
 	/* see if we should be clearing this entry */
 	if (value == NULL) {
 
-		rv = apr_memcache_delete(context->cache_memcache, key, 0);
+		rv = apr_memcache_delete(context->cache_memcache,
+				oidc_cache_memcache_get_key(r->pool, section, key), 0);
 
 		// TODO: error strings ?
 		if (rv != APR_SUCCESS) {
-			ap_log_rerror(APLOG_MARK, APLOG_ERR, rv, r,
-					"oidc_cache_memcache_set: apr_memcache_delete returned an error");
+			oidc_error(r, "apr_memcache_delete returned an error");
 		}
 
 	} else {
@@ -237,13 +245,13 @@ static apr_byte_t oidc_cache_memcache_set(request_rec *r, const char *key,
 		apr_uint32_t timeout = apr_time_sec(expiry - apr_time_now());
 
 		/* store it */
-		rv = apr_memcache_set(context->cache_memcache, key, (char *)value,
-				strlen(value), timeout, 0);
+		rv = apr_memcache_set(context->cache_memcache,
+				oidc_cache_memcache_get_key(r->pool, section, key),
+				(char *) value, strlen(value), timeout, 0);
 
 		// TODO: error strings ?
 		if (rv != APR_SUCCESS) {
-			ap_log_rerror(APLOG_MARK, APLOG_ERR, rv, r,
-					"oidc_cache_memcache_set: apr_memcache_set returned an error");
+			oidc_error(r, "apr_memcache_set returned an error");
 		}
 	}
 
