@@ -21,7 +21,6 @@
  * Copyright (C) 2013-2014 Ping Identity Corporation
  * All rights reserved.
  *
- * The contents of this file are the property of Ping Identity Corporation.
  * For further information please contact:
  *
  *      Ping Identity Corporation
@@ -73,8 +72,6 @@
 
 /* validate SSL server certificates by default */
 #define OIDC_DEFAULT_SSL_VALIDATE_SERVER 1
-/* default token endpoint authentication method */
-#define OIDC_DEFAULT_ENDPOINT_AUTH "client_secret_basic"
 /* default scope requested from the OP */
 #define OIDC_DEFAULT_SCOPE "openid"
 /* default claim delimiter for multi-valued claims passed in a HTTP header */
@@ -113,6 +110,10 @@
 #define OIDC_DEFAULT_IDTOKEN_IAT_SLACK 600
 /* for file-based caching: clean interval in seconds */
 #define OIDC_DEFAULT_CACHE_FILE_CLEAN_INTERVAL 60
+/* set httponly flag on cookies */
+#define OIDC_DEFAULT_COOKIE_HTTPONLY 1
+/* default cookie path */
+#define OIDC_DEFAULT_COOKIE_PATH "/"
 
 extern module AP_MODULE_DECLARE_DATA auth_openidc_module;
 
@@ -294,11 +295,6 @@ static const char *oidc_set_cache_type(cmd_parms *cmd, void *ptr,
 				arg));
 	}
 
-	cfg->cache_cfg =
-			cfg->cache->create_config ?
-					cfg->cache->create_config(cmd->server->process->pool) :
-					NULL;
-
 	return NULL;
 }
 
@@ -365,8 +361,8 @@ static const char *oidc_set_signed_response_alg(cmd_parms *cmd,
 	}
 
 	return apr_psprintf(cmd->pool, "parameter must be one of %s",
-			apr_array_pstrcat(cmd->pool, apr_jws_supported_algorithms(cmd->pool),
-					'|'));
+			apr_array_pstrcat(cmd->pool,
+					apr_jws_supported_algorithms(cmd->pool), '|'));
 }
 
 /*
@@ -382,8 +378,8 @@ static const char *oidc_set_encrypted_response_alg(cmd_parms *cmd,
 	}
 
 	return apr_psprintf(cmd->pool, "parameter must be one of %s",
-			apr_array_pstrcat(cmd->pool, apr_jwe_supported_algorithms(cmd->pool),
-					'|'));
+			apr_array_pstrcat(cmd->pool,
+					apr_jwe_supported_algorithms(cmd->pool), '|'));
 }
 
 /*
@@ -399,8 +395,8 @@ static const char *oidc_set_encrypted_response_enc(cmd_parms *cmd,
 	}
 
 	return apr_psprintf(cmd->pool, "parameter must be one of %s",
-			apr_array_pstrcat(cmd->pool, apr_jwe_supported_encryptions(cmd->pool),
-					'|'));
+			apr_array_pstrcat(cmd->pool,
+					apr_jwe_supported_encryptions(cmd->pool), '|'));
 }
 
 /*
@@ -434,33 +430,88 @@ static const char *oidc_set_session_inactivity_timeout(cmd_parms *cmd,
 /*
  * add a public key from an X.509 file to our list of JWKs with public keys
  */
-static const char *oidc_set_public_key_files_enc(cmd_parms *cmd, void *dummy, const char *arg) {
+static const char *oidc_set_public_key_files_enc(cmd_parms *cmd, void *dummy,
+		const char *arg) {
 	oidc_cfg *cfg = (oidc_cfg *) ap_get_module_config(
 			cmd->server->module_config, &auth_openidc_module);
 	char *jwk = NULL;
 	char *kid = NULL;
-	if (apr_jwk_x509_to_rsa_jwk(cmd->pool, arg, &jwk, &kid) == FALSE) {
-		return apr_psprintf(cmd->pool, "apr_jwk_x509_to_rsa_jwk failed for: %s", arg);
+	if (apr_jwk_pem_to_json(cmd->pool, arg, &jwk, &kid) == FALSE) {
+		return apr_psprintf(cmd->pool, "apr_jwk_pem_to_json failed for: %s",
+				arg);
 	}
-	if (cfg->public_keys == NULL) cfg->public_keys = apr_hash_make(cmd->pool);
+	if (cfg->public_keys == NULL)
+		cfg->public_keys = apr_hash_make(cmd->pool);
 	apr_hash_set(cfg->public_keys, kid, APR_HASH_KEY_STRING, jwk);
-    return NULL;
+	return NULL;
 }
 
 /*
  * add a private key from an RSA private key file to our list of JWKs with private keys
  */
-static const char *oidc_set_private_key_files_enc(cmd_parms *cmd, void *dummy, const char *arg) {
+static const char *oidc_set_private_key_files_enc(cmd_parms *cmd, void *dummy,
+		const char *arg) {
 	oidc_cfg *cfg = (oidc_cfg *) ap_get_module_config(
 			cmd->server->module_config, &auth_openidc_module);
 	char *jwk = NULL;
 	char *kid = NULL;
 	if (apr_jwk_private_key_to_rsa_jwk(cmd->pool, arg, &jwk, &kid) == FALSE) {
-		return apr_psprintf(cmd->pool, "apr_jwk_private_key_to_rsa_jwk failed for: %s", arg);
+		return apr_psprintf(cmd->pool,
+				"apr_jwk_private_key_to_rsa_jwk failed for: %s", arg);
 	}
-	if (cfg->private_keys == NULL) cfg->private_keys = apr_hash_make(cmd->pool);
+	if (cfg->private_keys == NULL)
+		cfg->private_keys = apr_hash_make(cmd->pool);
 	apr_hash_set(cfg->private_keys, kid, APR_HASH_KEY_STRING, jwk);
-    return NULL;
+	return NULL;
+}
+
+static int oidc_pass_idtoken_as_str2int(const char *v) {
+	if (apr_strnatcmp(v, "claims") == 0)
+		return OIDC_PASS_IDTOKEN_AS_CLAIMS;
+	if (apr_strnatcmp(v, "payload") == 0)
+		return OIDC_PASS_IDTOKEN_AS_PAYLOAD;
+	if (apr_strnatcmp(v, "serialized") == 0)
+		return OIDC_PASS_IDTOKEN_AS_SERIALIZED;
+	return -1;
+}
+
+/*
+ * define how to pass the id_token/claims in HTTP headers
+ */
+static const char * oidc_set_pass_idtoken_as(cmd_parms *cmd, void *dummy,
+		const char *v1, const char *v2, const char *v3) {
+	oidc_cfg *cfg = (oidc_cfg *) ap_get_module_config(
+			cmd->server->module_config, &auth_openidc_module);
+
+	int b = oidc_pass_idtoken_as_str2int(v1);
+	if (b != -1) {
+		cfg->pass_idtoken_as = b;
+	} else {
+		return apr_psprintf(cmd->pool, "Invalid value \"%s\" for directive %s",
+				v1, cmd->directive->directive);
+	}
+	if (v2) {
+		b = oidc_pass_idtoken_as_str2int(v2);
+		if (b != -1) {
+			cfg->pass_idtoken_as |= b;
+		} else {
+			return apr_psprintf(cmd->pool,
+					"Invalid value \"%s\" for directive %s", v2,
+					cmd->directive->directive);
+		}
+		if (v3) {
+			b = oidc_pass_idtoken_as_str2int(v3);
+			if (b != -1) {
+				cfg->pass_idtoken_as |= b;
+			} else {
+				return apr_psprintf(cmd->pool,
+						"Invalid value \"%s\" for directive %s", v3,
+						cmd->directive->directive);
+			}
+		}
+	}
+
+	return NULL;
 }
 
 /*
@@ -473,18 +524,23 @@ void *oidc_create_server_config(apr_pool_t *pool, server_rec *svr) {
 
 	c->redirect_uri = NULL;
 	c->discover_url = NULL;
-	c->default_url = NULL;
+	c->default_sso_url = NULL;
+	c->default_slo_url = NULL;
 	c->public_keys = NULL;
 	c->private_keys = NULL;
 
+	c->provider.metadata_url = NULL;
 	c->provider.issuer = NULL;
 	c->provider.authorization_endpoint_url = NULL;
 	c->provider.token_endpoint_url = NULL;
-	c->provider.token_endpoint_auth = OIDC_DEFAULT_ENDPOINT_AUTH;
+	c->provider.token_endpoint_auth = NULL;
+	c->provider.token_endpoint_params = NULL;
 	c->provider.userinfo_endpoint_url = NULL;
 	c->provider.client_id = NULL;
 	c->provider.client_secret = NULL;
 	c->provider.registration_endpoint_url = NULL;
+	c->provider.check_session_iframe = NULL;
+	c->provider.end_session_endpoint = NULL;
 	c->provider.jwks_uri = NULL;
 
 	c->provider.ssl_validate_server = OIDC_DEFAULT_SSL_VALIDATE_SERVER;
@@ -510,12 +566,12 @@ void *oidc_create_server_config(apr_pool_t *pool, server_rec *svr) {
 	c->oauth.client_id = NULL;
 	c->oauth.client_secret = NULL;
 	c->oauth.validate_endpoint_url = NULL;
-	c->oauth.validate_endpoint_auth = OIDC_DEFAULT_ENDPOINT_AUTH;
+	c->oauth.validate_endpoint_auth = NULL;
 	c->oauth.remote_user_claim = OIDC_DEFAULT_OAUTH_CLAIM_REMOTE_USER;
 
 	c->cache = &oidc_cache_shm;
-	c->cache_cfg =
-			c->cache->create_config ? c->cache->create_config(pool) : NULL;
+	c->cache_cfg = NULL;
+
 	c->cache_file_dir = NULL;
 	c->cache_file_clean_interval = OIDC_DEFAULT_CACHE_FILE_CLEAN_INTERVAL;
 	c->cache_memcache_servers = NULL;
@@ -533,6 +589,8 @@ void *oidc_create_server_config(apr_pool_t *pool, server_rec *svr) {
 	c->claim_delimiter = OIDC_DEFAULT_CLAIM_DELIMITER;
 	c->claim_prefix = OIDC_DEFAULT_CLAIM_PREFIX;
 	c->remote_user_claim = OIDC_DEFAULT_CLAIM_REMOTE_USER;
+	c->pass_idtoken_as = OIDC_PASS_IDTOKEN_AS_CLAIMS;
+	c->cookie_http_only = OIDC_DEFAULT_COOKIE_HTTPONLY;
 
 	c->outgoing_proxy = NULL;
 	c->crypto_passphrase = NULL;
@@ -556,13 +614,20 @@ void *oidc_merge_server_config(apr_pool_t *pool, void *BASE, void *ADD) {
 			add->redirect_uri != NULL ? add->redirect_uri : base->redirect_uri;
 	c->discover_url =
 			add->discover_url != NULL ? add->discover_url : base->discover_url;
-	c->default_url =
-			add->default_url != NULL ? add->default_url : base->default_url;
+	c->default_sso_url =
+			add->default_sso_url != NULL ?
+					add->default_sso_url : base->default_sso_url;
+	c->default_slo_url =
+			add->default_slo_url != NULL ?
+					add->default_slo_url : base->default_slo_url;
 	c->public_keys =
 			add->public_keys != NULL ? add->public_keys : base->public_keys;
 	c->private_keys =
 			add->private_keys != NULL ? add->private_keys : base->private_keys;
 
+	c->provider.metadata_url =
+			add->provider.metadata_url != NULL ?
+					add->provider.metadata_url : base->provider.metadata_url;
 	c->provider.issuer =
 			add->provider.issuer != NULL ?
 					add->provider.issuer : base->provider.issuer;
@@ -575,18 +640,20 @@ void *oidc_merge_server_config(apr_pool_t *pool, void *BASE, void *ADD) {
 					add->provider.token_endpoint_url :
 					base->provider.token_endpoint_url;
 	c->provider.token_endpoint_auth =
-			apr_strnatcmp(add->provider.token_endpoint_auth,
-			OIDC_DEFAULT_ENDPOINT_AUTH) != 0 ?
+			add->provider.token_endpoint_auth != NULL ?
 					add->provider.token_endpoint_auth :
 					base->provider.token_endpoint_auth;
+	c->provider.token_endpoint_params =
+			add->provider.token_endpoint_params != NULL ?
+					add->provider.token_endpoint_params :
+					base->provider.token_endpoint_params;
 	c->provider.userinfo_endpoint_url =
 			add->provider.userinfo_endpoint_url != NULL ?
 					add->provider.userinfo_endpoint_url :
 					base->provider.userinfo_endpoint_url;
 	c->provider.jwks_uri =
 			add->provider.jwks_uri != NULL ?
-					add->provider.jwks_uri :
-					base->provider.jwks_uri;
+					add->provider.jwks_uri : base->provider.jwks_uri;
 	c->provider.client_id =
 			add->provider.client_id != NULL ?
 					add->provider.client_id : base->provider.client_id;
@@ -597,6 +664,14 @@ void *oidc_merge_server_config(apr_pool_t *pool, void *BASE, void *ADD) {
 			add->provider.registration_endpoint_url != NULL ?
 					add->provider.registration_endpoint_url :
 					base->provider.registration_endpoint_url;
+	c->provider.check_session_iframe =
+			add->provider.check_session_iframe != NULL ?
+					add->provider.check_session_iframe :
+					base->provider.check_session_iframe;
+	c->provider.end_session_endpoint =
+			add->provider.end_session_endpoint != NULL ?
+					add->provider.end_session_endpoint :
+					base->provider.end_session_endpoint;
 
 	c->provider.ssl_validate_server =
 			add->provider.ssl_validate_server
@@ -621,7 +696,7 @@ void *oidc_merge_server_config(apr_pool_t *pool, void *BASE, void *ADD) {
 	c->provider.response_type =
 			apr_strnatcmp(add->provider.response_type,
 					OIDC_DEFAULT_RESPONSE_TYPE) != 0 ?
-					add->provider.response_type : base->provider.response_type;
+							add->provider.response_type : base->provider.response_type;
 	c->provider.response_mode =
 			add->provider.response_mode != NULL ?
 					add->provider.response_mode : base->provider.response_mode;
@@ -683,15 +758,14 @@ void *oidc_merge_server_config(apr_pool_t *pool, void *BASE, void *ADD) {
 					add->oauth.validate_endpoint_url :
 					base->oauth.validate_endpoint_url;
 	c->oauth.validate_endpoint_auth =
-			apr_strnatcmp(add->oauth.validate_endpoint_auth,
-			OIDC_DEFAULT_ENDPOINT_AUTH) != 0 ?
+			add->oauth.validate_endpoint_auth != NULL ?
 					add->oauth.validate_endpoint_auth :
 					base->oauth.validate_endpoint_auth;
 	c->oauth.remote_user_claim =
 			apr_strnatcmp(add->oauth.remote_user_claim,
 					OIDC_DEFAULT_OAUTH_CLAIM_REMOTE_USER) != 0 ?
-					add->oauth.remote_user_claim :
-					base->oauth.remote_user_claim;
+							add->oauth.remote_user_claim :
+							base->oauth.remote_user_claim;
 
 	c->http_timeout_long =
 			add->http_timeout_long != OIDC_DEFAULT_HTTP_TIMEOUT_LONG ?
@@ -710,11 +784,10 @@ void *oidc_merge_server_config(apr_pool_t *pool, void *BASE, void *ADD) {
 
 	if (add->cache != &oidc_cache_shm) {
 		c->cache = add->cache;
-		c->cache_cfg = add->cache_cfg;
 	} else {
 		c->cache = base->cache;
-		c->cache_cfg = base->cache_cfg;
 	}
+	c->cache_cfg = NULL;
 
 	c->cache_file_dir =
 			add->cache_file_dir != NULL ?
@@ -750,7 +823,13 @@ void *oidc_merge_server_config(apr_pool_t *pool, void *BASE, void *ADD) {
 	c->remote_user_claim =
 			apr_strnatcmp(add->remote_user_claim,
 					OIDC_DEFAULT_CLAIM_REMOTE_USER) != 0 ?
-					add->remote_user_claim : base->remote_user_claim;
+							add->remote_user_claim : base->remote_user_claim;
+	c->pass_idtoken_as =
+			add->pass_idtoken_as != OIDC_PASS_IDTOKEN_AS_CLAIMS ?
+					add->pass_idtoken_as : base->pass_idtoken_as;
+	c->cookie_http_only =
+			add->cookie_http_only != OIDC_DEFAULT_COOKIE_HTTPONLY ?
+					add->cookie_http_only : base->cookie_http_only;
 
 	c->outgoing_proxy =
 			add->outgoing_proxy != NULL ?
@@ -773,7 +852,7 @@ void *oidc_merge_server_config(apr_pool_t *pool, void *BASE, void *ADD) {
 void *oidc_create_dir_config(apr_pool_t *pool, char *path) {
 	oidc_dir_cfg *c = apr_pcalloc(pool, sizeof(oidc_dir_cfg));
 	c->cookie = OIDC_DEFAULT_COOKIE;
-	c->cookie_path = NULL;
+	c->cookie_path = OIDC_DEFAULT_COOKIE_PATH;
 	c->authn_header = OIDC_DEFAULT_AUTHN_HEADER;
 	return (c);
 }
@@ -789,7 +868,8 @@ void *oidc_merge_dir_config(apr_pool_t *pool, void *BASE, void *ADD) {
 			apr_strnatcasecmp(add->cookie, OIDC_DEFAULT_COOKIE) != 0 ?
 					add->cookie : base->cookie);
 	c->cookie_path = (
-			add->cookie_path != NULL ? add->cookie_path : base->cookie_path);
+			apr_strnatcasecmp(add->cookie_path, OIDC_DEFAULT_COOKIE_PATH) != 0 ?
+					add->cookie_path : base->cookie_path);
 	c->authn_header = (
 			add->authn_header != OIDC_DEFAULT_AUTHN_HEADER ?
 					add->authn_header : base->authn_header);
@@ -800,9 +880,7 @@ void *oidc_merge_dir_config(apr_pool_t *pool, void *BASE, void *ADD) {
  * report a config error
  */
 static int oidc_check_config_error(server_rec *s, const char *config_str) {
-	ap_log_error(APLOG_MARK, APLOG_ERR, 0, s,
-			"oidc_check_config_error: mandatory parameter '%s' is not set",
-			config_str);
+	oidc_serror(s, "mandatory parameter '%s' is not set", config_str);
 	return HTTP_INTERNAL_SERVER_ERROR;
 }
 
@@ -811,9 +889,12 @@ static int oidc_check_config_error(server_rec *s, const char *config_str) {
  */
 static int oidc_check_config_openid_openidc(server_rec *s, oidc_cfg *c) {
 
-	if ((c->metadata_dir == NULL) && (c->provider.issuer == NULL)) {
-		ap_log_error(APLOG_MARK, APLOG_ERR, 0, s,
-				"oidc_check_config_openid_openidc: one of 'OIDCProviderIssuer' or 'OIDCMetadataDir' must be set");
+	apr_uri_t r_uri;
+
+	if ((c->metadata_dir == NULL) && (c->provider.issuer == NULL)
+			&& (c->provider.metadata_url == NULL)) {
+		oidc_serror(s,
+				"one of 'OIDCProviderIssuer', 'OIDCProviderMetadataURL' or 'OIDCMetadataDir' must be set");
 		return HTTP_INTERNAL_SERVER_ERROR;
 	}
 
@@ -823,34 +904,48 @@ static int oidc_check_config_openid_openidc(server_rec *s, oidc_cfg *c) {
 		return oidc_check_config_error(s, "OIDCCryptoPassphrase");
 
 	if (c->metadata_dir == NULL) {
-		if (c->provider.issuer == NULL)
-			return oidc_check_config_error(s, "OIDCProviderIssuer");
-		if (c->provider.authorization_endpoint_url == NULL)
-			return oidc_check_config_error(s,
-					"OIDCProviderAuthorizationEndpoint");
-		// TODO: this depends on the configured OIDCResponseType now
-		if (c->provider.token_endpoint_url == NULL)
-			return oidc_check_config_error(s, "OIDCProviderTokenEndpoint");
+		if (c->provider.metadata_url == NULL) {
+			if (c->provider.issuer == NULL)
+				return oidc_check_config_error(s, "OIDCProviderIssuer");
+			if (c->provider.authorization_endpoint_url == NULL)
+				return oidc_check_config_error(s,
+						"OIDCProviderAuthorizationEndpoint");
+			// TODO: this depends on the configured OIDCResponseType now
+			//			if (c->provider.token_endpoint_url == NULL)
+			//				return oidc_check_config_error(s, "OIDCProviderTokenEndpoint");
+		} else {
+			apr_uri_parse(s->process->pconf, c->provider.metadata_url, &r_uri);
+			if (apr_strnatcmp(r_uri.scheme, "http") == 0) {
+				oidc_swarn(s,
+						"the URL scheme (%s) of the configured OIDCProviderMetadataURL SHOULD be \"https\" for security reasons!",
+						r_uri.scheme);
+			}
+		}
 		if (c->provider.client_id == NULL)
 			return oidc_check_config_error(s, "OIDCClientID");
 		// TODO: this depends on the configured OIDCResponseType now
 		if (c->provider.client_secret == NULL)
 			return oidc_check_config_error(s, "OIDCClientSecret");
+	} else {
+		if (c->provider.metadata_url != NULL) {
+			oidc_serror(s,
+					"only one of 'OIDCProviderMetadataURL' or 'OIDCMetadataDir' should be set");
+			return HTTP_INTERNAL_SERVER_ERROR;
+		}
 	}
 
-	apr_uri_t r_uri;
 	apr_uri_parse(s->process->pconf, c->redirect_uri, &r_uri);
 	if (apr_strnatcmp(r_uri.scheme, "https") != 0) {
-		ap_log_error(APLOG_MARK, APLOG_WARNING, 0, s,
-				"oidc_check_config_openid_openidc: the URL scheme (%s) of the configured OIDCRedirectURI SHOULD be \"https\" for security reasons (moreover: some Providers may reject non-HTTPS URLs)",
+		oidc_swarn(s,
+				"the URL scheme (%s) of the configured OIDCRedirectURI SHOULD be \"https\" for security reasons (moreover: some Providers may reject non-HTTPS URLs)",
 				r_uri.scheme);
 	}
 
 	if (c->cookie_domain != NULL) {
 		char *p = strstr(r_uri.hostname, c->cookie_domain);
 		if ((p == NULL) || (apr_strnatcmp(c->cookie_domain, p) != 0)) {
-			ap_log_error(APLOG_MARK, APLOG_ERR, 0, s,
-					"oidc_check_config_openid_openidc: the domain (%s) configured in OIDCCookieDomain does not match the URL hostname (%s) of the configured OIDCRedirectURI (%s): setting \"state\" and \"session\" cookies will not work!",
+			oidc_serror(s,
+					"the domain (%s) configured in OIDCCookieDomain does not match the URL hostname (%s) of the configured OIDCRedirectURI (%s): setting \"state\" and \"session\" cookies will not work!",
 					c->cookie_domain, r_uri.hostname, c->redirect_uri);
 			return HTTP_INTERNAL_SERVER_ERROR;
 		}
@@ -883,8 +978,7 @@ static int oidc_config_check_vhost_config(apr_pool_t *pool, server_rec *s) {
 	oidc_cfg *cfg = ap_get_module_config(s->module_config,
 			&auth_openidc_module);
 
-	ap_log_error(APLOG_MARK, OIDC_DEBUG, 0, s,
-			"oidc_config_check_vhost_config: entering");
+	oidc_sdebug(s, "enter");
 
 	if ((cfg->metadata_dir != NULL) || (cfg->provider.issuer == NULL)
 			|| (cfg->redirect_uri != NULL)
@@ -966,13 +1060,14 @@ static void oidc_ssl_id_callback(CRYPTO_THREADID *id) {
 
 static apr_status_t oidc_cleanup(void *data) {
 
-	server_rec *sp = (server_rec *)data;
+	server_rec *sp = (server_rec *) data;
 	while (sp != NULL) {
 		oidc_cfg *cfg = (oidc_cfg *) ap_get_module_config(sp->module_config,
 				&auth_openidc_module);
+		oidc_crypto_destroy(cfg, sp);
 		if (cfg->cache->destroy != NULL) {
 			if (cfg->cache->destroy(sp) != APR_SUCCESS) {
-				ap_log_error(APLOG_MARK, APLOG_ERR, 0, sp, "oidc_cleanup: cache destroy function failed");
+				oidc_serror(sp, "cache destroy function failed");
 			}
 		}
 		sp = sp->next;
@@ -983,7 +1078,7 @@ static apr_status_t oidc_cleanup(void *data) {
 		CRYPTO_set_locking_callback(NULL);
 #ifdef OPENSSL_NO_THREADID
 	if (CRYPTO_get_id_callback() == oidc_ssl_id_callback)
-	CRYPTO_set_id_callback(NULL);
+		CRYPTO_set_id_callback(NULL);
 #else
 	if (CRYPTO_THREADID_get_callback() == oidc_ssl_id_callback)
 		CRYPTO_THREADID_set_callback(NULL);
@@ -992,6 +1087,10 @@ static apr_status_t oidc_cleanup(void *data) {
 #endif /* defined(OPENSSL_THREADS) && APR_HAS_THREADS */
 	EVP_cleanup();
 	curl_global_cleanup();
+
+	ap_log_error(APLOG_MARK, APLOG_INFO, 0, (server_rec *) data,
+			"%s - shutdown", NAMEVERSION);
+
 	return APR_SUCCESS;
 }
 
@@ -1016,6 +1115,8 @@ static int oidc_post_config(apr_pool_t *pool, apr_pool_t *p1, apr_pool_t *p2,
 		return OK;
 	}
 
+	ap_log_error(APLOG_MARK, APLOG_INFO, 0, s, "%s - init", NAMEVERSION);
+
 	curl_global_init(CURL_GLOBAL_ALL);
 	OpenSSL_add_all_digests();
 
@@ -1078,8 +1179,8 @@ static int oidc_post_config(apr_pool_t *pool, apr_pool_t *p1, apr_pool_t *p2,
 
 #if MODULE_MAGIC_NUMBER_MAJOR >= 20100714
 static const authz_provider authz_oidc_provider = {
-	&oidc_authz_checker,
-	NULL,
+		&oidc_authz_checker,
+		NULL,
 };
 #endif
 
@@ -1092,7 +1193,7 @@ static void oidc_child_init(apr_pool_t *p, server_rec *s) {
 				&auth_openidc_module);
 		if (cfg->cache->child_init != NULL) {
 			if (cfg->cache->child_init(p, s) != APR_SUCCESS) {
-				ap_log_error(APLOG_MARK, APLOG_ERR, 0, s, "oidc_child_init: cfg->cache->child_init failed");
+				oidc_serror(s, "cfg->cache->child_init failed");
 			}
 		}
 		s = s->next;
@@ -1120,9 +1221,14 @@ void oidc_register_hooks(apr_pool_t *pool) {
  */
 const command_rec oidc_config_cmds[] = {
 
+		AP_INIT_TAKE1("OIDCProviderMetadataURL", oidc_set_string_slot,
+				(void*)APR_OFFSETOF(oidc_cfg, provider.metadata_url),
+				RSRC_CONF,
+				"OpenID Connect OP configuration metadata URL."),
 		AP_INIT_TAKE1("OIDCProviderIssuer", oidc_set_string_slot,
 				(void*)APR_OFFSETOF(oidc_cfg, provider.issuer),
-				RSRC_CONF, "OpenID Connect OP issuer identifier."),
+				RSRC_CONF,
+				"OpenID Connect OP issuer identifier."),
 		AP_INIT_TAKE1("OIDCProviderAuthorizationEndpoint",
 				oidc_set_https_slot,
 				(void *)APR_OFFSETOF(oidc_cfg, provider.authorization_endpoint_url),
@@ -1138,16 +1244,31 @@ const command_rec oidc_config_cmds[] = {
 				(void *)APR_OFFSETOF(oidc_cfg, provider.token_endpoint_auth),
 				RSRC_CONF,
 				"Specify an authentication method for the OpenID OP Token Endpoint (e.g.: client_secret_basic)"),
+		AP_INIT_TAKE1("OIDCProviderTokenEndpointParams",
+				oidc_set_string_slot,
+				(void *)APR_OFFSETOF(oidc_cfg, provider.token_endpoint_params),
+				RSRC_CONF,
+				"Define extra parameters that will be posted to the OpenID OP Token Endpoint (e.g.: param1=value1&param2=value2, all urlencoded)."),
 		AP_INIT_TAKE1("OIDCProviderUserInfoEndpoint",
 				oidc_set_https_slot,
 				(void *)APR_OFFSETOF(oidc_cfg, provider.userinfo_endpoint_url),
 				RSRC_CONF,
 				"Define the OpenID OP UserInfo Endpoint URL (e.g.: https://localhost:9031/idp/userinfo.openid)"),
+		AP_INIT_TAKE1("OIDCProviderCheckSessionIFrame",
+				oidc_set_url_slot,
+				(void *)APR_OFFSETOF(oidc_cfg, provider.check_session_iframe),
+				RSRC_CONF,
+				"Define the OpenID OP Check Session iFrame URL."),
+		AP_INIT_TAKE1("OIDCProviderEndSessionEndpoint",
+				oidc_set_url_slot,
+				(void *)APR_OFFSETOF(oidc_cfg, provider.end_session_endpoint),
+				RSRC_CONF,
+				"Define the OpenID OP End Session Endpoint URL."),
 		AP_INIT_TAKE1("OIDCProviderJwksUri",
 				oidc_set_https_slot,
 				(void *)APR_OFFSETOF(oidc_cfg, provider.jwks_uri),
 				RSRC_CONF,
-				"Define the OpenID OP JWKS URL (e.g.: https://macbook:9031/pf/JWKS)"),
+				"Define the OpenID OP JWKS URL (e.g.: https://localhost:9031/pf/JWKS)"),
 		AP_INIT_TAKE1("OIDCResponseType",
 				oidc_set_response_type,
 				(void *)APR_OFFSETOF(oidc_cfg, provider.response_type),
@@ -1169,12 +1290,12 @@ const command_rec oidc_config_cmds[] = {
 				"The fully qualified names of the files that contain the RSA private keys that can be used to decrypt content sent to us by the OP."),
 
 		AP_INIT_TAKE1("OIDCClientJwksUri",
-				oidc_set_signed_response_alg,
+				oidc_set_https_slot,
 				(void *)APR_OFFSETOF(oidc_cfg, provider.client_jwks_uri),
 				RSRC_CONF,
 				"Define the Client JWKS URL (e.g.: https://localhost/protected/?jwks=rsa)"),
 		AP_INIT_TAKE1("OIDCIDTokenSignedResponseAlg",
-				oidc_set_https_slot,
+				oidc_set_signed_response_alg,
 				(void *)APR_OFFSETOF(oidc_cfg, provider.id_token_signed_response_alg),
 				RSRC_CONF,
 				"The algorithm that the OP should use to sign the id_token (used only in dynamic client registration); must be one of [RS256|RS384|RS512|PS256|PS384|PS512|HS256|HS384|HS512]"),
@@ -1252,13 +1373,22 @@ const command_rec oidc_config_cmds[] = {
 				(void *)APR_OFFSETOF(oidc_cfg, discover_url),
 				RSRC_CONF,
 				"Defines an external IDP Discovery page"),
-		AP_INIT_TAKE1("OIDCDefaultURL", oidc_set_string_slot,
-				(void *)APR_OFFSETOF(oidc_cfg, default_url),
+		AP_INIT_TAKE1("OIDCDefaultURL", oidc_set_url_slot,
+				(void *)APR_OFFSETOF(oidc_cfg, default_sso_url),
 				RSRC_CONF,
 				"Defines the default URL where the user is directed to in case of 3rd-party initiated SSO."),
+		AP_INIT_TAKE1("OIDCDefaultLoggedOutURL", oidc_set_url_slot,
+				(void *)APR_OFFSETOF(oidc_cfg, default_slo_url),
+				RSRC_CONF,
+				"Defines the default URL where the user is directed to after logout."),
 		AP_INIT_TAKE1("OIDCCookieDomain",
 				oidc_set_cookie_domain, NULL, RSRC_CONF,
 				"Specify domain element for OIDC session cookie."),
+		AP_INIT_FLAG("OIDCCookieHTTPOnly",
+				oidc_set_flag_slot,
+				(void *) APR_OFFSETOF(oidc_cfg, cookie_http_only),
+				RSRC_CONF,
+				"Defines whether or not the cookie httponly flag is set on cookies."),
 		AP_INIT_TAKE1("OIDCOutgoingProxy",
 				oidc_set_string_slot,
 				(void*)APR_OFFSETOF(oidc_cfg, outgoing_proxy),
@@ -1283,6 +1413,11 @@ const command_rec oidc_config_cmds[] = {
 				(void*)APR_OFFSETOF(oidc_cfg, remote_user_claim),
 				RSRC_CONF,
 				"The claim that is used when setting the REMOTE_USER variable for OpenID Connect protected paths."),
+		AP_INIT_TAKE123("OIDCPassIDTokenAs",
+				oidc_set_pass_idtoken_as,
+				NULL,
+				RSRC_CONF,
+				"The format in which the id_token is passed in (a) header(s); must be one or more of: claims|payload|serialized"),
 
 		AP_INIT_TAKE1("OIDCOAuthClientID", oidc_set_string_slot,
 				(void*)APR_OFFSETOF(oidc_cfg, oauth.client_id),
