@@ -21,7 +21,6 @@
  * Copyright (C) 2013-2014 Ping Identity Corporation
  * All rights reserved.
  *
- * The contents of this file are the property of Ping Identity Corporation.
  * For further information please contact:
  *
  *      Ping Identity Corporation
@@ -76,10 +75,44 @@ APLOG_USE_MODULE(auth_openidc);
 #define OIDC_DEBUG APLOG_DEBUG
 #endif
 
+#define oidc_log(r, level, fmt, ...) ap_log_rerror(APLOG_MARK, level, 0, r,"%s: %s", __FUNCTION__, apr_psprintf(r->pool, fmt, ##__VA_ARGS__))
+#define oidc_slog(s, level, fmt, ...) ap_log_error(APLOG_MARK, level, 0, s, "%s: %s", __FUNCTION__, apr_psprintf(s->process->pool, fmt, ##__VA_ARGS__))
+
+#define oidc_debug(r, fmt, ...) oidc_log(r, OIDC_DEBUG, fmt, ##__VA_ARGS__)
+#define oidc_warn(r, fmt, ...) oidc_log(r, APLOG_WARNING, fmt, ##__VA_ARGS__)
+#define oidc_error(r, fmt, ...) oidc_log(r, APLOG_ERR, fmt, ##__VA_ARGS__)
+
+#define oidc_sdebug(s, fmt, ...) oidc_slog(s, OIDC_DEBUG, fmt, ##__VA_ARGS__)
+#define oidc_swarn(s, fmt, ...) oidc_slog(s, APLOG_WARNING, fmt, ##__VA_ARGS__)
+#define oidc_serror(s, fmt, ...) oidc_slog(s, APLOG_ERR, fmt, ##__VA_ARGS__)
+
+#ifndef NAMEVER
+#define NAMEVERSION "mod_auth_openidc-0.0.0"
+#else
+#define STRINGIFY(x) #x
+#define TOSTRING(x) STRINGIFY(x)
+#define NAMEVERSION TOSTRING(NAMEVER)
+#endif
+
 /* key for storing the claims in the session context */
 #define OIDC_CLAIMS_SESSION_KEY "claims"
 /* key for storing the id_token in the session context */
+#define OIDC_IDTOKEN_CLAIMS_SESSION_KEY "id_token_claims"
+/* key for storing the raw id_token in the session context */
 #define OIDC_IDTOKEN_SESSION_KEY "id_token"
+/* key for storing the access_token in the session context */
+#define OIDC_ACCESSTOKEN_SESSION_KEY "access_token"
+
+/* key for storing the session_state in the session context */
+#define OIDC_SESSION_STATE_SESSION_KEY "session_state"
+/* key for storing the issuer in the session context */
+#define OIDC_ISSUER_SESSION_KEY "issuer"
+/* key for storing the client_id in the session context */
+#define OIDC_CLIENTID_SESSION_KEY "client_id"
+/* key for storing the check_session_iframe in the session context */
+#define OIDC_CHECK_IFRAME_SESSION_KEY "check_session_iframe"
+/* key for storing the end_session_endpoint in the session context */
+#define OIDC_LOGOUT_ENDPOINT_SESSION_KEY "end_session_endpoint"
 
 /* parameter name of the callback URL in the discovery response */
 #define OIDC_DISC_CB_PARAM "oidc_callback"
@@ -89,38 +122,55 @@ APLOG_USE_MODULE(auth_openidc);
 #define OIDC_DISC_RT_PARAM "target_link_uri"
 /* parameter name of login hint in the discovery response */
 #define OIDC_DISC_LH_PARAM "login_hint"
+/* parameter name of parameters that need to be passed in the authentication request */
+#define OIDC_DISC_AR_PARAM "auth_request_params"
 
 /* value that indicates to use server-side cache based session tracking */
 #define OIDC_SESSION_TYPE_22_SERVER_CACHE 0
 /* value that indicates to use client cookie based session tracking */
 #define OIDC_SESSION_TYPE_22_CLIENT_COOKIE 1
 
-/* name of the cookie that binds the state in the authorization request/response to the browser */
-#define OIDCStateCookieName  "mod_auth_openidc_state"
+/* pass id_token as individual claims in headers (default) */
+#define OIDC_PASS_IDTOKEN_AS_CLAIMS     1
+/* pass id_token payload as JSON object in header*/
+#define OIDC_PASS_IDTOKEN_AS_PAYLOAD    2
+/* pass id_token in compact serialized format in header*/
+#define OIDC_PASS_IDTOKEN_AS_SERIALIZED 4
+
+/* prefix of the cookie that binds the state in the authorization request/response to the browser */
+#define OIDCStateCookiePrefix  "mod_auth_openidc_state_"
 
 /* the (global) key for the mod_auth_openidc related state that is stored in the request userdata context */
 #define OIDC_USERDATA_KEY "mod_auth_openidc_state"
 
-/* use for plain GET in HTTP calls to endpoints */
-#define OIDC_HTTP_GET 0
-/* use for url-form-encoded HTTP POST calls to endpoints */
-#define OIDC_HTTP_POST_FORM 1
-/* use for JSON encoded POST calls to endpoints */
-#define OIDC_HTTP_POST_JSON 2
-
 /* input filter hook name */
 #define OIDC_UTIL_HTTP_SENDSTRING "OIDC_UTIL_HTTP_SENDSTRING"
 
 /* the name of the keyword that follows the Require primitive to indicate claims-based authorization */
 #define OIDC_REQUIRE_NAME "claim"
 
+/* defines for how long provider metadata will be cached */
+#define OIDC_CACHE_PROVIDER_METADATA_EXPIRY_DEFAULT 86400
+
+/* cache sections */
+#define OIDC_CACHE_SECTION_JTI "jti"
+#define OIDC_CACHE_SECTION_SESSION "session"
+#define OIDC_CACHE_SECTION_NONCE "nonce"
+#define OIDC_CACHE_SECTION_JWKS "jwks"
+#define OIDC_CACHE_SECTION_ACCESS_TOKEN "access_token"
+#define OIDC_CACHE_SECTION_PROVIDER "provider"
+
 typedef struct oidc_provider_t {
+	char *metadata_url;
 	char *issuer;
 	char *authorization_endpoint_url;
 	char *token_endpoint_url;
 	char *token_endpoint_auth;
+	char *token_endpoint_params;
 	char *userinfo_endpoint_url;
 	char *registration_endpoint_url;
+	char *check_session_iframe;
+	char *end_session_endpoint;
 	char *jwks_uri;
 	char *client_id;
 	char *client_secret;
@@ -164,7 +214,9 @@ typedef struct oidc_cfg {
 	/* (optional) external OP discovery page */
 	char *discover_url;
 	/* (optional) default URL for 3rd-party initiated SSO */
-	char *default_url;
+	char *default_sso_url;
+	/* (optional) default URL to go to after logout */
+	char *default_slo_url;
 
 	/* public keys in JWK format, used by parters for encrypting JWTs sent to us */
 	apr_hash_t *public_keys;
@@ -206,6 +258,8 @@ typedef struct oidc_cfg {
 	char *claim_delimiter;
 	char *claim_prefix;
 	char *remote_user_claim;
+	int pass_idtoken_as;
+	int cookie_http_only;
 
 	char *outgoing_proxy;
 
@@ -242,17 +296,18 @@ typedef struct oidc_proto_state {
 	const char *issuer;
 	const char *response_type;
 	const char *response_mode;
+	const char *prompt;
 	apr_time_t timestamp;
 } oidc_proto_state;
 
-int oidc_proto_authorization_request(request_rec *r, struct oidc_provider_t *provider, const char *login_hint, const char *redirect_uri, const char *state, oidc_proto_state *proto_state);
+int oidc_proto_authorization_request(request_rec *r, struct oidc_provider_t *provider, const char *login_hint, const char *redirect_uri, const char *state, oidc_proto_state *proto_state, const char *id_token_hint, const char *auth_request_params);
 apr_byte_t oidc_proto_is_post_authorization_response(request_rec *r, oidc_cfg *cfg);
 apr_byte_t oidc_proto_is_redirect_authorization_response(request_rec *r, oidc_cfg *cfg);
 apr_byte_t oidc_proto_check_token_type(request_rec *r, oidc_provider_t *provider, const char *token_type);
 apr_byte_t oidc_proto_resolve_code(request_rec *r, oidc_cfg *cfg, oidc_provider_t *provider, const char *code, char **s_id_token, char **s_access_token, char **s_token_type);
 apr_byte_t oidc_proto_resolve_userinfo(request_rec *r, oidc_cfg *cfg, oidc_provider_t *provider, const char *access_token, const char **response, json_t **claims);
 apr_byte_t oidc_proto_account_based_discovery(request_rec *r, oidc_cfg *cfg, const char *acct, char **issuer);
-apr_byte_t oidc_proto_parse_idtoken(request_rec *r, oidc_cfg *cfg, oidc_provider_t *provider, const char *id_token, const char *nonce, char **user, apr_jwt_t **jwt);
+apr_byte_t oidc_proto_parse_idtoken(request_rec *r, oidc_cfg *cfg, oidc_provider_t *provider, const char *id_token, const char *nonce, char **user, apr_jwt_t **jwt, apr_byte_t is_code_flow);
 apr_byte_t oidc_proto_validate_access_token(request_rec *r, oidc_provider_t *provider, apr_jwt_t *jwt, const char *response_type, const char *access_token, const char *token_type);
 apr_byte_t oidc_proto_validate_code(request_rec *r, oidc_provider_t *provider, apr_jwt_t *jwt, const char *response_type, const char *code);
 int oidc_proto_javascript_implicit(request_rec *r, oidc_cfg *c);
@@ -279,22 +334,25 @@ void oidc_register_hooks(apr_pool_t *pool);
 
 // oidc_util.c
 int oidc_strnenvcmp(const char *a, const char *b, int len);
-int oidc_base64url_encode(request_rec *r, char **dst, const char *src, int src_len);
-int oidc_base64url_decode(request_rec *r, char **dst, const char *src, int padding);
+int oidc_base64url_encode(request_rec *r, char **dst, const char *src, int src_len, int remove_padding);
+int oidc_base64url_decode(request_rec *r, char **dst, const char *src, int add_padding);
 int oidc_encrypt_base64url_encode_string(request_rec *r, char **dst, const char *src);
 int oidc_base64url_decode_decrypt_string(request_rec *r, char **dst, const char *src);
 char *oidc_get_current_url(const request_rec *r, const oidc_cfg *c);
 char *oidc_url_encode(const request_rec *r, const char *str, const char *charsToEncode);
 char *oidc_normalize_header_name(const request_rec *r, const char *str);
 
-void oidc_util_set_cookie(request_rec *r, const char *cookieName, const char *cookieValue);
-char *oidc_util_get_cookie(request_rec *r, char *cookieName);
-apr_byte_t oidc_util_http_call(request_rec *r, const char *url, int action, const apr_table_t *params, const char *basic_auth, const char *bearer_token, int ssl_validate_server, const char **response, int timeout, const char *outgoing_proxy);
+void oidc_util_set_cookie(request_rec *r, const char *cookieName, const char *cookieValue, apr_time_t expires);
+char *oidc_util_get_cookie(request_rec *r, const char *cookieName);
+apr_byte_t oidc_util_http_get(request_rec *r, const char *url, const apr_table_t *params, const char *basic_auth, const char *bearer_token, int ssl_validate_server, const char **response, int timeout, const char *outgoing_proxy);
+apr_byte_t oidc_util_http_post_form(request_rec *r, const char *url, const apr_table_t *params, const char *basic_auth, const char *bearer_token, int ssl_validate_server, const char **response, int timeout, const char *outgoing_proxy);
+apr_byte_t oidc_util_http_post_json(request_rec *r, const char *url, const json_t *data, const char *basic_auth, const char *bearer_token, int ssl_validate_server, const char **response, int timeout, const char *outgoing_proxy);
 apr_byte_t oidc_util_request_matches_url(request_rec *r, const char *url);
 apr_byte_t oidc_util_request_has_parameter(request_rec *r, const char* param);
 apr_byte_t oidc_util_get_request_parameter(request_rec *r, char *name, char **value);
 apr_byte_t oidc_util_decode_json_and_check_error(request_rec *r, const char *str, json_t **json);
-int oidc_util_http_sendstring(request_rec *r, const char *html, int success_rvalue);
+int oidc_util_http_send(request_rec *r, const char *data, int data_len, const char *content_type, int success_rvalue);
+int oidc_util_html_send(request_rec *r, const char *html, int success_rvalue);
 char *oidc_util_escape_string(const request_rec *r, const char *str);
 char *oidc_util_unescape_string(const request_rec *r, const char *str);
 apr_byte_t oidc_util_read_post(request_rec *r, apr_table_t *table);
@@ -303,18 +361,23 @@ apr_byte_t oidc_util_file_read(request_rec *r, const char *path, char **result);
 apr_byte_t oidc_util_issuer_match(const char *a, const char *b);
 int oidc_util_html_send_error(request_rec *r, const char *error, const char *description, int status_code);
 apr_byte_t oidc_util_json_array_has_value(request_rec *r, json_t *haystack, const char *needle);
+void oidc_util_set_app_header(request_rec *r, const char *s_key, const char *s_value, const char *claim_prefix);
 void oidc_util_set_app_headers(request_rec *r, const json_t *j_attrs, const char *claim_prefix, const char *claim_delimiter);
 apr_hash_t *oidc_util_spaced_string_to_hashtable(apr_pool_t *pool, const char *str);
 apr_byte_t oidc_util_spaced_string_equals(apr_pool_t *pool, const char *a, const char *b);
 apr_byte_t oidc_util_spaced_string_contains(apr_pool_t *pool, const char *response_type, const char *match);
 apr_byte_t oidc_json_object_get_string(apr_pool_t *pool, json_t *json, const char *name, char **value, const char *default_value);
 apr_byte_t oidc_json_object_get_int(apr_pool_t *pool, json_t *json, const char *name, int *value, const int default_value);
+char *oidc_util_html_escape(apr_pool_t *pool, const char *input);
 
 // oidc_crypto.c
 unsigned char *oidc_crypto_aes_encrypt(request_rec *r, oidc_cfg *cfg, unsigned char *plaintext, int *len);
 unsigned char *oidc_crypto_aes_decrypt(request_rec *r, oidc_cfg *cfg, unsigned char *ciphertext, int *len);
+apr_byte_t oidc_crypto_destroy(oidc_cfg *cfg, server_rec *s);
 
 // oidc_metadata.c
+apr_byte_t oidc_metadata_provider_retrieve(request_rec *r, oidc_cfg *cfg, const char *issuer, const char *url, json_t **j_metadata, const char **response);
+apr_byte_t oidc_metadata_provider_parse(request_rec *r, json_t *j_provider, oidc_provider_t *provider);
 apr_byte_t oidc_metadata_list(request_rec *r, oidc_cfg *cfg, apr_array_header_t **arr);
 apr_byte_t oidc_metadata_get(request_rec *r, oidc_cfg *cfg, const char *selected, oidc_provider_t **provider);
 apr_byte_t oidc_metadata_jwks_get(request_rec *r, oidc_cfg *cfg, oidc_provider_t *provider, json_t **j_jwks, apr_byte_t *refresh);
