@@ -21,7 +21,6 @@
  * Copyright (C) 2013-2014 Ping Identity Corporation
  * All rights reserved.
  *
- * The contents of this file are the property of Ping Identity Corporation.
  * For further information please contact:
  *
  *      Ping Identity Corporation
@@ -70,11 +69,9 @@ extern module AP_MODULE_DECLARE_DATA auth_openidc_module;
  * base64url encode a string
  */
 int oidc_base64url_encode(request_rec *r, char **dst, const char *src,
-		int src_len) {
-	// TODO: always padded now, do we need an option to remove the padding?
+		int src_len, int remove_padding) {
 	if ((src == NULL) || (src_len <= 0)) {
-		ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, r,
-				"oidc_base64url_encode: not encoding anything; src=NULL and/or src_len<1");
+		oidc_error(r, "not encoding anything; src=NULL and/or src_len<1");
 		return -1;
 	}
 	int enc_len = apr_base64_encode_len(src_len);
@@ -90,6 +87,15 @@ int oidc_base64url_encode(request_rec *r, char **dst, const char *src,
 			enc[i] = ',';
 		i++;
 	}
+	if (remove_padding) {
+		/* remove /0 and padding */
+		enc_len--;
+		if (enc[enc_len - 1] == ',')
+			enc_len--;
+		if (enc[enc_len - 1] == ',')
+			enc_len--;
+		enc[enc_len] = '\0';
+	}
 	*dst = enc;
 	return enc_len;
 }
@@ -98,11 +104,9 @@ int oidc_base64url_encode(request_rec *r, char **dst, const char *src,
  * base64url decode a string
  */
 int oidc_base64url_decode(request_rec *r, char **dst, const char *src,
-		int padding) {
-	// TODO: check base64url decoding/encoding code and look for alternatives?
+		int add_padding) {
 	if (src == NULL) {
-		ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, r,
-				"oidc_base64url_decode: not decoding anything; src=NULL");
+		oidc_error(r, "not decoding anything; src=NULL");
 		return -1;
 	}
 	char *dec = apr_pstrdup(r->pool, src);
@@ -116,7 +120,7 @@ int oidc_base64url_decode(request_rec *r, char **dst, const char *src,
 			dec[i] = '=';
 		i++;
 	}
-	if (padding == 1) {
+	if (add_padding == 1) {
 		switch (strlen(dec) % 4) {
 		case 0:
 			break;
@@ -146,11 +150,10 @@ int oidc_encrypt_base64url_encode_string(request_rec *r, char **dst,
 	unsigned char *crypted = oidc_crypto_aes_encrypt(r, c,
 			(unsigned char *) src, &crypted_len);
 	if (crypted == NULL) {
-		ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, r,
-				"oidc_encrypt_base64url_encode_string: oidc_crypto_aes_encrypt failed");
+		oidc_error(r, "oidc_crypto_aes_encrypt failed");
 		return -1;
 	}
-	return oidc_base64url_encode(r, dst, (const char *) crypted, crypted_len);
+	return oidc_base64url_encode(r, dst, (const char *) crypted, crypted_len, 1);
 }
 
 /*
@@ -161,17 +164,15 @@ int oidc_base64url_decode_decrypt_string(request_rec *r, char **dst,
 	oidc_cfg *c = ap_get_module_config(r->server->module_config,
 			&auth_openidc_module);
 	char *decbuf = NULL;
-	int dec_len = oidc_base64url_decode(r, &decbuf, src, 0);
+	int dec_len = oidc_base64url_decode(r, &decbuf, src, 1);
 	if (dec_len <= 0) {
-		ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, r,
-				"oidc_base64url_decode_decrypt_string: oidc_base64url_decode failed");
+		oidc_error(r, "oidc_base64url_decode failed");
 		return -1;
 	}
 	*dst = (char *) oidc_crypto_aes_decrypt(r, c, (unsigned char *) decbuf,
 			&dec_len);
 	if (*dst == NULL) {
-		ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, r,
-				"oidc_base64url_decode_decrypt_string: oidc_crypto_aes_decrypt failed");
+		oidc_error(r, "oidc_crypto_aes_decrypt failed");
 		return -1;
 	}
 	return dec_len;
@@ -230,14 +231,12 @@ int oidc_strnenvcmp(const char *a, const char *b, int len) {
 char *oidc_util_escape_string(const request_rec *r, const char *str) {
 	CURL *curl = curl_easy_init();
 	if (curl == NULL) {
-		ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, r,
-				"oidc_escape_string: curl_easy_init() error");
+		oidc_error(r, "curl_easy_init() error");
 		return NULL;
 	}
 	char *result = curl_easy_escape(curl, str, 0);
 	if (result == NULL) {
-		ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, r,
-				"oidc_escape_string: curl_easy_escape() error");
+		oidc_error(r, "curl_easy_escape() error");
 		return NULL;
 	}
 	char *rv = apr_pstrdup(r->pool, result);
@@ -252,23 +251,51 @@ char *oidc_util_escape_string(const request_rec *r, const char *str) {
 char *oidc_util_unescape_string(const request_rec *r, const char *str) {
 	CURL *curl = curl_easy_init();
 	if (curl == NULL) {
-		ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, r,
-				"oidc_util_unescape_string: curl_easy_init() error");
+		oidc_error(r, "curl_easy_init() error");
 		return NULL;
 	}
 	char *result = curl_easy_unescape(curl, str, 0, 0);
 	if (result == NULL) {
-		ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, r,
-				"oidc_util_unescape_string: curl_easy_unescape() error");
+		oidc_error(r, "curl_easy_unescape() error");
 		return NULL;
 	}
 	char *rv = apr_pstrdup(r->pool, result);
 	curl_free(result);
 	curl_easy_cleanup(curl);
-	//ap_log_rerror(APLOG_MARK, OIDC_DEBUG, 0, r, "oidc_util_unescape_string: input=\"%s\", output=\"%s\"", str, rv);
+	//oidc_debug(r, "input=\"%s\", output=\"%s\"", str, rv);
 	return rv;
 }
 
+/*
+ * HTML escape a string
+ */
+char *oidc_util_html_escape(apr_pool_t *pool, const char *s) {
+	const char chars[6] = { '&', '\'', '\"', '>', '<', '\0' };
+	const char * const replace[] =
+	{ "&amp;", "&apos;", "&quot;", "&gt;", "&lt;", };
+	unsigned int i, j = 0, k, n = 0, len = strlen(chars);
+	int m = 0;
+	char *r = apr_pcalloc(pool, strlen(s) * 6);
+	for (i = 0; i < strlen(s); i++) {
+		for (n = 0; n < len; n++) {
+			if (s[i] == chars[n]) {
+				m = strlen(replace[n]);
+				for (k = 0; k < m; k++)
+					r[j + k] = replace[n][k];
+				j += m;
+				break;
+			}
+		}
+		if (n == len) {
+			r[j] = s[i];
+			j++;
+		}
+	}
+	r[j] = '\0';
+	return apr_pstrdup(pool, r);
+}
+
+
 /*
  * get the URL scheme that is currently being accessed
  */
@@ -289,7 +316,8 @@ static const char *oidc_get_current_url_scheme(const request_rec *r) {
 /*
  * get the URL port that is currently being accessed
  */
-static const char *oidc_get_current_url_port(const request_rec *r, const oidc_cfg *c, const char *scheme_str) {
+static const char *oidc_get_current_url_port(const request_rec *r,
+		const oidc_cfg *c, const char *scheme_str) {
 	/* first see if there's a proxy/load-balancer in front of us */
 	const char *port_str = apr_table_get(r->headers_in, "X-Forwarded-Port");
 	if (port_str == NULL) {
@@ -328,11 +356,11 @@ char *oidc_get_current_url(const request_rec *r, const oidc_cfg *c) {
 	if (p != NULL)
 		*p = '\0';
 
-	char *url = apr_pstrcat(r->pool, scheme_str, "://", host_str, port_str, r->uri,
-			(r->args != NULL && *r->args != '\0' ? "?" : ""), r->args, NULL);
+	char *url = apr_pstrcat(r->pool, scheme_str, "://", host_str, port_str,
+			r->uri, (r->args != NULL && *r->args != '\0' ? "?" : ""), r->args,
+			NULL);
 
-	ap_log_rerror(APLOG_MARK, OIDC_DEBUG, 0, r,
-			"oidc_get_current_url: current URL '%s'", url);
+	oidc_debug(r, "current URL '%s'", url);
 
 	return url;
 }
@@ -372,62 +400,35 @@ typedef struct oidc_http_encode_t {
  */
 static int oidc_http_add_form_url_encoded_param(void* rec, const char* key,
 		const char* value) {
-	// TODO: handle arrays of strings?
 	oidc_http_encode_t *ctx = (oidc_http_encode_t*) rec;
 	const char *sep = apr_strnatcmp(ctx->encoded_params, "") == 0 ? "" : "&";
-	ap_log_rerror(APLOG_MARK, OIDC_DEBUG, 0, ctx->r,
-			"oidc_http_add_post_param: adding parameter: %s=%s to %s (sep=%s)",
-			key, value, ctx->encoded_params, sep);
 	ctx->encoded_params = apr_psprintf(ctx->r->pool, "%s%s%s=%s",
 			ctx->encoded_params, sep, oidc_util_escape_string(ctx->r, key),
 			oidc_util_escape_string(ctx->r, value));
 	return 1;
 }
 
-/*
- * add a JSON name/value pair
- */
-static int oidc_http_add_json_param(void* rec, const char* key,
-		const char* value) {
-	oidc_http_encode_t *ctx = (oidc_http_encode_t*) rec;
-	const char *sep = apr_strnatcmp(ctx->encoded_params, "") == 0 ? "" : ",";
-	ap_log_rerror(APLOG_MARK, OIDC_DEBUG, 0, ctx->r,
-			"oidc_http_add_json_param: adding parameter: %s=%s to %s", key,
-			value, ctx->encoded_params);
-	if ( (value[0] == '[') || (value[0] == '{') ) {
-		// TODO hacky hacky, we need an array so we already encoded it :-)
-		ctx->encoded_params = apr_psprintf(ctx->r->pool, "%s%s\"%s\" : %s",
-				ctx->encoded_params, sep, key, value);
-	} else {
-		ctx->encoded_params = apr_psprintf(ctx->r->pool, "%s%s\"%s\": \"%s\"",
-				ctx->encoded_params, sep, key, value);
-	}
-	return 1;
-}
-
 /*
  * execute a HTTP (GET or POST) request
  */
-apr_byte_t oidc_util_http_call(request_rec *r, const char *url, int action,
-		const apr_table_t *params, const char *basic_auth,
+static apr_byte_t oidc_util_http_call(request_rec *r, const char *url,
+		const char *data, const char *content_type, const char *basic_auth,
 		const char *bearer_token, int ssl_validate_server,
 		const char **response, int timeout, const char *outgoing_proxy) {
 	char curlError[CURL_ERROR_SIZE];
 	oidc_curl_buffer curlBuffer;
 	CURL *curl;
 	struct curl_slist *h_list = NULL;
-	int nr_of_params = (params != NULL) ? apr_table_elts(params)->nelts : 0;
 
 	/* do some logging about the inputs */
-	ap_log_rerror(APLOG_MARK, OIDC_DEBUG, 0, r,
-			"oidc_util_http_call: entering, url=%s, action=%d, #params=%d, basic_auth=%s, bearer_token=%s, ssl_validate_server=%d",
-			url, action, nr_of_params, basic_auth, bearer_token,
+	oidc_debug(r,
+			"url=%s, data=%s, content_type=%s, basic_auth=%s, bearer_token=%s, ssl_validate_server=%d",
+			url, data, content_type, basic_auth, bearer_token,
 			ssl_validate_server);
 
 	curl = curl_easy_init();
 	if (curl == NULL) {
-		ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, r,
-				"oidc_util_http_call: curl_easy_init() error");
+		oidc_error(r, "curl_easy_init() error");
 		return FALSE;
 	}
 
@@ -481,65 +482,17 @@ apr_byte_t oidc_util_http_call(request_rec *r, const char *url, int action,
 		curl_easy_setopt(curl, CURLOPT_USERPWD, basic_auth);
 	}
 
-	/* the POST contents */
-	oidc_http_encode_t data = { r, "" };
-
-	if (action == OIDC_HTTP_POST_JSON) {
-
-		/* POST JSON data */
-
-		if (nr_of_params > 0) {
-
-			/* add the parameters in JSON formatting */
-			apr_table_do(oidc_http_add_json_param, &data, params, NULL);
-			/* surround it by brackets to make it a valid JSON object */
-			data.encoded_params = apr_psprintf(r->pool, "{ %s }",
-					data.encoded_params);
-
-			/* set the data and log the event */
-			ap_log_rerror(APLOG_MARK, OIDC_DEBUG, 0, r,
-					"oidc_util_http_call: setting JSON parameters: %s",
-					data.encoded_params);
-			curl_easy_setopt(curl, CURLOPT_POSTFIELDS, data.encoded_params);
-		}
-
+	if (data != NULL) {
+		/* set POST data */
+		curl_easy_setopt(curl, CURLOPT_POSTFIELDS, data);
 		/* set HTTP method to POST */
 		curl_easy_setopt(curl, CURLOPT_POST, 1);
+	}
 
-		/* and overwrite the default url-form-encoded content-type */
-//		h_list = curl_slist_append(h_list,
-//				"Content-type: application/json; charset=UTF-8");
-		h_list = curl_slist_append(h_list, "Content-type: application/json");
-
-	} else if (action == OIDC_HTTP_POST_FORM) {
-
-		/* POST url-form-encoded data */
-
-		if (nr_of_params > 0) {
-
-			apr_table_do(oidc_http_add_form_url_encoded_param, &data, params,
-					NULL);
-
-			ap_log_rerror(APLOG_MARK, OIDC_DEBUG, 0, r,
-					"oidc_util_http_call: setting post parameters: %s",
-					data.encoded_params);
-			curl_easy_setopt(curl, CURLOPT_POSTFIELDS, data.encoded_params);
-		} // else: probably should warn here...
-
-		/* CURLOPT_POST needed at least to set: Content-Type: application/x-www-form-urlencoded */
-		curl_easy_setopt(curl, CURLOPT_POST, 1);
-
-	} else if (nr_of_params > 0) {
-
-		/* HTTP GET with #params > 0 */
-
-		apr_table_do(oidc_http_add_form_url_encoded_param, &data, params, NULL);
-		const char *sep = strchr(url, '?') != NULL ? "&" : "?";
-		url = apr_psprintf(r->pool, "%s%s%s", url, sep, data.encoded_params);
-
-		/* log that the URL has changed now */
-		ap_log_rerror(APLOG_MARK, OIDC_DEBUG, 0, r,
-				"oidc_util_http_call: added query parameters to URL: %s", url);
+	if (content_type != NULL) {
+		/* set content type */
+		h_list = curl_slist_append(h_list,
+				apr_psprintf(r->pool, "Content-type: %s", content_type));
 	}
 
 	/* see if we need to add any custom headers */
@@ -552,9 +505,7 @@ apr_byte_t oidc_util_http_call(request_rec *r, const char *url, int action,
 	/* call it and record the result */
 	int rv = TRUE;
 	if (curl_easy_perform(curl) != CURLE_OK) {
-		ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, r,
-				"oidc_util_http_call: curl_easy_perform() failed on: %s (%s)",
-				url, curlError);
+		oidc_error(r, "curl_easy_perform() failed on: %s (%s)", url, curlError);
 		rv = FALSE;
 		goto out;
 	}
@@ -562,8 +513,7 @@ apr_byte_t oidc_util_http_call(request_rec *r, const char *url, int action,
 	*response = apr_pstrndup(r->pool, curlBuffer.buf, curlBuffer.written);
 
 	/* set and log the response */
-	ap_log_rerror(APLOG_MARK, OIDC_DEBUG, 0, r,
-			"oidc_util_http_call: response=%s", *response);
+	oidc_debug(r, "response=%s", *response);
 
 	out:
 
@@ -575,6 +525,68 @@ apr_byte_t oidc_util_http_call(request_rec *r, const char *url, int action,
 	return rv;
 }
 
+/*
+ * execute HTTP GET request
+ */
+apr_byte_t oidc_util_http_get(request_rec *r, const char *url,
+		const apr_table_t *params, const char *basic_auth,
+		const char *bearer_token, int ssl_validate_server,
+		const char **response, int timeout, const char *outgoing_proxy) {
+
+	if ((params != NULL) && (apr_table_elts(params)->nelts > 0)) {
+		oidc_http_encode_t data = { r, "" };
+		apr_table_do(oidc_http_add_form_url_encoded_param, &data, params, NULL);
+		const char *sep = strchr(url, '?') != NULL ? "&" : "?";
+		url = apr_psprintf(r->pool, "%s%s%s", url, sep, data.encoded_params);
+		oidc_debug(r, "get URL=\"%s\"", url);
+	}
+
+	return oidc_util_http_call(r, url, NULL, NULL, basic_auth, bearer_token,
+			ssl_validate_server, response, timeout, outgoing_proxy);
+}
+
+/*
+ * execute HTTP POST request with form-encoded data
+ */
+apr_byte_t oidc_util_http_post_form(request_rec *r, const char *url,
+		const apr_table_t *params, const char *basic_auth,
+		const char *bearer_token, int ssl_validate_server,
+		const char **response, int timeout, const char *outgoing_proxy) {
+
+	const char *data = NULL;
+	if ((params != NULL) && (apr_table_elts(params)->nelts > 0)) {
+		oidc_http_encode_t encode_data = { r, "" };
+		apr_table_do(oidc_http_add_form_url_encoded_param, &encode_data, params,
+				NULL);
+		data = encode_data.encoded_params;
+		oidc_debug(r, "post data=\"%s\"", data);
+	}
+
+	return oidc_util_http_call(r, url, data,
+			"application/x-www-form-urlencoded", basic_auth, bearer_token,
+			ssl_validate_server, response, timeout, outgoing_proxy);
+}
+
+/*
+ * execute HTTP POST request with JSON-encoded data
+ */
+apr_byte_t oidc_util_http_post_json(request_rec *r, const char *url,
+		const json_t *json, const char *basic_auth, const char *bearer_token,
+		int ssl_validate_server, const char **response, int timeout,
+		const char *outgoing_proxy) {
+
+	char *data = NULL;
+	if (json != NULL) {
+		char *s_value = json_dumps(json, 0);
+		data = apr_pstrdup(r->pool, s_value);
+		free(s_value);
+	}
+
+	return oidc_util_http_call(r, url, data, "application/json", basic_auth,
+			bearer_token, ssl_validate_server, response, timeout,
+			outgoing_proxy);
+}
+
 /*
  * get the current path from the request in a normalized way
  */
@@ -601,8 +613,8 @@ static char *oidc_util_get_cookie_path(request_rec *r) {
 		if (strncmp(d->cookie_path, requestPath, strlen(d->cookie_path)) == 0)
 			rv = d->cookie_path;
 		else {
-			ap_log_rerror(APLOG_MARK, APLOG_WARNING, 0, r,
-					"oidc_util_get_cookie_path: OIDCCookiePath (%s) not a substring of request path, using request path (%s) for cookie",
+			oidc_warn(r,
+					"OIDCCookiePath (%s) not a substring of request path, using request path (%s) for cookie",
 					d->cookie_path, requestPath);
 			rv = requestPath;
 		}
@@ -616,24 +628,34 @@ static char *oidc_util_get_cookie_path(request_rec *r) {
  * set a cookie in the HTTP response headers
  */
 void oidc_util_set_cookie(request_rec *r, const char *cookieName,
-		const char *cookieValue) {
+		const char *cookieValue, apr_time_t expires) {
 
 	oidc_cfg *c = ap_get_module_config(r->server->module_config,
 			&auth_openidc_module);
-	char *headerString, *currentCookies;
-	/* construct the cookie value */
-	headerString = apr_psprintf(r->pool, "%s=%s;%s;Path=%s%s",
-			cookieName,
-			cookieValue,
-			((apr_strnatcasecmp("https", oidc_get_current_url_scheme(r)) == 0) ? ";Secure" : ""),
-			oidc_util_get_cookie_path(r),
-			c->cookie_domain != NULL ? apr_psprintf(r->pool, ";Domain=%s", c->cookie_domain) : ""
-		);
+	char *headerString, *currentCookies, *expiresString = NULL;
 
 	/* see if we need to clear the cookie */
 	if (apr_strnatcmp(cookieValue, "") == 0)
-		headerString = apr_psprintf(r->pool, "%s;expires=0;Max-Age=0",
-				headerString);
+		expires = 0;
+
+	/* construct the expire value */
+	if (expires != -1) {
+		expiresString = (char *) apr_pcalloc(r->pool, APR_RFC822_DATE_LEN);
+		if (apr_rfc822_date(expiresString, expires) != APR_SUCCESS) {
+			oidc_error(r, "could not set cookie expiry date");
+		}
+	}
+
+	/* construct the cookie value */
+	headerString = apr_psprintf(r->pool, "%s=%s;%s;Path=%s%s%s%s", cookieName,
+			cookieValue,
+			((apr_strnatcasecmp("https", oidc_get_current_url_scheme(r)) == 0) ?
+					";Secure" : ""), oidc_util_get_cookie_path(r),
+			c->cookie_domain != NULL ?
+					apr_psprintf(r->pool, ";Domain=%s", c->cookie_domain) : "",
+			c->cookie_http_only != FALSE ? ";HttpOnly" : "",
+			(expiresString == NULL) ?
+					"" : apr_psprintf(r->pool, "; expires=%s", expiresString));
 
 	/* use r->err_headers_out so we always print our headers (even on 302 redirect) - headers_out only prints on 2xx responses */
 	apr_table_add(r->err_headers_out, "Set-Cookie", headerString);
@@ -647,15 +669,13 @@ void oidc_util_set_cookie(request_rec *r, const char *cookieName,
 				(apr_pstrcat(r->pool, headerString, ";", currentCookies, NULL)));
 
 	/* do some logging */
-	ap_log_rerror(APLOG_MARK, OIDC_DEBUG, 0, r,
-			"oidc_util_set_cookie: adding outgoing header: Set-Cookie: %s",
-			headerString);
+	oidc_debug(r, "adding outgoing header: Set-Cookie: %s", headerString);
 }
 
 /*
  * get a cookie from the HTTP request
  */
-char *oidc_util_get_cookie(request_rec *r, char *cookieName) {
+char *oidc_util_get_cookie(request_rec *r, const char *cookieName) {
 	char *cookie, *tokenizerCtx, *rv = NULL;
 
 	/* get the Cookie value */
@@ -689,8 +709,7 @@ char *oidc_util_get_cookie(request_rec *r, char *cookieName) {
 	}
 
 	/* log what we've found */
-	ap_log_rerror(APLOG_MARK, OIDC_DEBUG, 0, r, "oidc_get_cookie: returning %s",
-			rv);
+	oidc_debug(r, "returning %s", rv);
 
 	return rv;
 }
@@ -727,12 +746,12 @@ char *oidc_normalize_header_name(const request_rec *r, const char *str) {
  */
 apr_byte_t oidc_util_request_matches_url(request_rec *r, const char *url) {
 	apr_uri_t uri;
+	memset(&uri, 0, sizeof(apr_uri_t));
 	apr_uri_parse(r->pool, url, &uri);
 	apr_byte_t rc =
 			(apr_strnatcmp(r->parsed_uri.path, uri.path) == 0) ? TRUE : FALSE;
-	ap_log_rerror(APLOG_MARK, OIDC_DEBUG, 0, r,
-			"oidc_request_matches_url: comparing \"%s\"==\"%s\" (%d)",
-			r->parsed_uri.path, uri.path, rc);
+	oidc_debug(r, "comparing \"%s\"==\"%s\" (%d)", r->parsed_uri.path, uri.path,
+			rc);
 	return rc;
 }
 
@@ -781,16 +800,16 @@ apr_byte_t oidc_util_get_request_parameter(request_rec *r, char *name,
 /*
  * printout a JSON string value
  */
-static apr_byte_t oidc_util_json_string_print(request_rec *r,
-		json_t *result, const char *key, const char *log) {
+static apr_byte_t oidc_util_json_string_print(request_rec *r, json_t *result,
+		const char *key, const char *log) {
 	json_t *value = json_object_get(result, key);
 	if (value != NULL) {
 		if (json_is_string(value)) {
-			ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, r,
+			oidc_error(r,
 					"%s: response contained a \"%s\" key with string value: \"%s\"",
 					log, key, json_string_value(value));
 		} else {
-			ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, r,
+			oidc_error(r,
 					"%s: response contained an \"%s\" key but no string value",
 					log, key);
 		}
@@ -802,8 +821,7 @@ static apr_byte_t oidc_util_json_string_print(request_rec *r,
 /*
  * check a JSON object for "error" results and printout
  */
-static apr_byte_t oidc_util_check_json_error(request_rec *r,
-		json_t *json) {
+static apr_byte_t oidc_util_check_json_error(request_rec *r, json_t *json) {
 	if (oidc_util_json_string_print(r, json, "error",
 			"oidc_util_check_json_error") == TRUE) {
 		oidc_util_json_string_print(r, json, "error_description",
@@ -825,15 +843,13 @@ apr_byte_t oidc_util_decode_json_and_check_error(request_rec *r,
 	/* decode the JSON contents of the buffer */
 	if (*json == NULL) {
 		/* something went wrong */
-		ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, r,
-				"oidc_util_check_json_error: JSON parsing returned an error: %s", json_error.text);
+		oidc_error(r, "JSON parsing returned an error: %s", json_error.text);
 		return FALSE;
 	}
 
 	if (!json_is_object(*json)) {
 		/* oops, no JSON */
-		ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, r,
-				"oidc_util_check_json_error: parsed JSON did not contain a JSON object");
+		oidc_error(r, "parsed JSON did not contain a JSON object");
 		json_decref(*json);
 		*json = NULL;
 		return FALSE;
@@ -850,14 +866,14 @@ apr_byte_t oidc_util_decode_json_and_check_error(request_rec *r,
 }
 
 /*
- * sends HTML content to the user agent
+ * sends content to the user agent
  */
-int oidc_util_http_sendstring(request_rec *r, const char *html,
+int oidc_util_http_send(request_rec *r, const char *data, int data_len, const char *content_type,
 		int success_rvalue) {
-	ap_set_content_type(r, "text/html");
+	ap_set_content_type(r, content_type);
 	apr_bucket_brigade *bb = apr_brigade_create(r->pool,
 			r->connection->bucket_alloc);
-	apr_bucket *b = apr_bucket_transient_create(html, strlen(html),
+	apr_bucket *b = apr_bucket_transient_create(data, data_len,
 			r->connection->bucket_alloc);
 	APR_BRIGADE_INSERT_TAIL(bb, b);
 	b = apr_bucket_eos_create(r->connection->bucket_alloc);
@@ -868,6 +884,14 @@ int oidc_util_http_sendstring(request_rec *r, const char *html,
 	return success_rvalue;
 }
 
+/*
+ * sends HTML content to the user agent
+ */
+int oidc_util_html_send(request_rec *r, const char *html,
+		int success_rvalue) {
+	return oidc_util_http_send(r, html, strlen(html), "text/html", success_rvalue);
+}
+
 /*
  * read all bytes from the HTTP request
  */
@@ -956,8 +980,7 @@ apr_byte_t oidc_util_file_read(request_rec *r, const char *path, char **result)
 	/* open the file if it exists */
 	if ((rc = apr_file_open(&fd, path, APR_FOPEN_READ | APR_FOPEN_BUFFERED,
 	APR_OS_DEFAULT, r->pool)) != APR_SUCCESS) {
-		ap_log_rerror(APLOG_MARK, APLOG_WARNING, 0, r,
-				"oidc_util_file_read: no file found at: \"%s\"", path);
+		oidc_warn(r, "no file found at: \"%s\"", path);
 		return FALSE;
 	}
 
@@ -970,8 +993,7 @@ apr_byte_t oidc_util_file_read(request_rec *r, const char *path, char **result)
 
 	/* get the file info so we know its size */
 	if ((rc = apr_file_info_get(&finfo, APR_FINFO_SIZE, fd)) != APR_SUCCESS) {
-		ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, r,
-				"oidc_util_file_read: error calling apr_file_info_get on file: \"%s\" (%s)",
+		oidc_error(r, "error calling apr_file_info_get on file: \"%s\" (%s)",
 				path, apr_strerror(rc, s_err, sizeof(s_err)));
 		goto error_close;
 	}
@@ -983,9 +1005,8 @@ apr_byte_t oidc_util_file_read(request_rec *r, const char *path, char **result)
 	apr_size_t bytes_read = 0;
 	if ((rc = apr_file_read_full(fd, *result, finfo.size, &bytes_read))
 			!= APR_SUCCESS) {
-		ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, r,
-				"oidc_util_file_read: apr_file_read_full on (%s) returned an error: %s",
-				path, apr_strerror(rc, s_err, sizeof(s_err)));
+		oidc_error(r, "apr_file_read_full on (%s) returned an error: %s", path,
+				apr_strerror(rc, s_err, sizeof(s_err)));
 		goto error_close;
 	}
 
@@ -994,8 +1015,8 @@ apr_byte_t oidc_util_file_read(request_rec *r, const char *path, char **result)
 
 	/* check that we've got all of it */
 	if (bytes_read != finfo.size) {
-		ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, r,
-				"oidc_util_file_read: apr_file_read_full on (%s) returned less bytes (%" APR_SIZE_T_FMT ") than expected: (%" APR_OFF_T_FMT ")",
+		oidc_error(r,
+				"apr_file_read_full on (%s) returned less bytes (%" APR_SIZE_T_FMT ") than expected: (%" APR_OFF_T_FMT ")",
 				path, bytes_read, finfo.size);
 		goto error_close;
 	}
@@ -1005,18 +1026,16 @@ apr_byte_t oidc_util_file_read(request_rec *r, const char *path, char **result)
 	apr_file_close(fd);
 
 	/* log successful content retrieval */
-	ap_log_rerror(APLOG_MARK, OIDC_DEBUG, 0, r,
-			"oidc_util_file_read: file read successfully \"%s\"", path);
+	oidc_debug(r, "file read successfully \"%s\"", path);
 
 	return TRUE;
 
-error_close:
+	error_close:
 
 	apr_file_unlock(fd);
 	apr_file_close(fd);
 
-	ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, r,
-			"oidc_util_file_read: returning error");
+	oidc_error(r, "return error");
 
 	return FALSE;
 }
@@ -1047,28 +1066,28 @@ apr_byte_t oidc_util_issuer_match(const char *a, const char *b) {
  */
 int oidc_util_html_send_error(request_rec *r, const char *error,
 		const char *description, int status_code) {
-	char *msg = "<p>the OpenID Connect Provider returned an error:</p><p>";
+	char *msg =
+			"<html><body><p>the OpenID Connect Provider returned an error:</p>";
 
 	if (error != NULL) {
 		msg = apr_psprintf(r->pool, "%s<p>Error: <pre>%s</pre></p>", msg,
-				error);
+				oidc_util_html_escape(r->pool, error));
 	}
 	if (description != NULL) {
 		msg = apr_psprintf(r->pool, "%s<p>Description: <pre>%s</pre></p>", msg,
-				description);
+				oidc_util_html_escape(r->pool, description));
 	}
 
-	return oidc_util_http_sendstring(r, msg, status_code);
+	msg = apr_psprintf(r->pool, "%s</body></html>", msg);
+
+	return oidc_util_html_send(r, msg, status_code);
 }
 
 /*
  * see if a certain string value is part of a JSON array with string elements
  */
-apr_byte_t oidc_util_json_array_has_value(request_rec *r,
-		json_t *haystack, const char *needle) {
-
-	ap_log_rerror(APLOG_MARK, OIDC_DEBUG, 0, r,
-			"oidc_util_json_array_has_value: entering (%s)", needle);
+apr_byte_t oidc_util_json_array_has_value(request_rec *r, json_t *haystack,
+		const char *needle) {
 
 	if ((haystack == NULL) || (!json_is_array(haystack)))
 		return FALSE;
@@ -1077,8 +1096,8 @@ apr_byte_t oidc_util_json_array_has_value(request_rec *r,
 	for (i = 0; i < json_array_size(haystack); i++) {
 		json_t *elem = json_array_get(haystack, i);
 		if (!json_is_string(elem)) {
-			ap_log_rerror(APLOG_MARK, APLOG_ERR, 0, r,
-					"oidc_util_json_array_has_value: unhandled in-array JSON non-string object type [%d]", elem->type);
+			oidc_error(r, "unhandled in-array JSON non-string object type [%d]",
+					elem->type);
 			continue;
 		}
 		if (strcmp(json_string_value(elem), needle) == 0) {
@@ -1086,8 +1105,8 @@ apr_byte_t oidc_util_json_array_has_value(request_rec *r,
 		}
 	}
 
-//	ap_log_rerror(APLOG_MARK, OIDC_DEBUG, 0, r,
-//			"oidc_util_json_array_has_value: returning (%d=%d)", i,
+//	oidc_debug(r,
+//			"returning (%d=%d)", i,
 //			haystack->value.array->nelts);
 
 	return (i == json_array_size(haystack)) ? FALSE : TRUE;
@@ -1096,7 +1115,7 @@ apr_byte_t oidc_util_json_array_has_value(request_rec *r,
 /*
  * set an HTTP header to pass information to the application
  */
-static void oidc_util_set_app_header(request_rec *r, const char *s_key,
+void oidc_util_set_app_header(request_rec *r, const char *s_key,
 		const char *s_value, const char *claim_prefix) {
 
 	/* construct the header name, cq. put the prefix in front of a normalized key name */
@@ -1104,9 +1123,7 @@ static void oidc_util_set_app_header(request_rec *r, const char *s_key,
 			oidc_normalize_header_name(r, s_key));
 
 	/* do some logging about this event */
-	ap_log_rerror(APLOG_MARK, OIDC_DEBUG, 0, r,
-			"oidc_util_set_app_header: setting header \"%s: %s\"", s_name,
-			s_value);
+	oidc_debug(r, "setting header \"%s: %s\"", s_name, s_value);
 
 	/* now set the actual header name/value */
 	apr_table_set(r->headers_in, s_name, s_value);
@@ -1115,8 +1132,8 @@ static void oidc_util_set_app_header(request_rec *r, const char *s_key,
 /*
  * set the user/claims information from the session in HTTP headers passed on to the application
  */
-void oidc_util_set_app_headers(request_rec *r, const json_t *j_attrs, const char *claim_prefix,
-		const char *claim_delimiter) {
+void oidc_util_set_app_headers(request_rec *r, const json_t *j_attrs,
+		const char *claim_prefix, const char *claim_delimiter) {
 
 	char s_int[255];
 	json_t *j_value = NULL;
@@ -1124,19 +1141,22 @@ void oidc_util_set_app_headers(request_rec *r, const json_t *j_attrs, const char
 
 	/* if not attributes are set, nothing needs to be done */
 	if (j_attrs == NULL) {
-		ap_log_rerror(APLOG_MARK, OIDC_DEBUG, 0, r,
-				"oidc_util_set_app_headers: no attributes to set");
+		oidc_debug(r, "no attributes to set");
 		return;
 	}
 
 	/* loop over the claims in the JSON structure */
-	void *iter = json_object_iter((json_t*)j_attrs);
+	void *iter = json_object_iter((json_t*) j_attrs);
 	while (iter) {
 
 		/* get the next key/value entry */
 		s_key = json_object_iter_key(iter);
 		j_value = json_object_iter_value(iter);
 
+//		char *s_value= json_dumps(j_value, JSON_ENCODE_ANY);
+//		oidc_util_set_app_header(r, s_key, s_value, claim_prefix);
+//		free(s_value);
+
 		/* check if it is a single value string */
 		if (json_is_string(j_value)) {
 
@@ -1148,18 +1168,17 @@ void oidc_util_set_app_headers(request_rec *r, const json_t *j_attrs, const char
 
 			/* set boolean value in the application header whose name is based on the key and the prefix */
 			oidc_util_set_app_header(r, s_key,
-					json_is_true(j_value) ? "1" : "0", claim_prefix);
+			json_is_true(j_value) ? "1" : "0", claim_prefix);
 
 		} else if (json_is_integer(j_value)) {
 
-			if (sprintf(s_int, "%" JSON_INTEGER_FORMAT, json_integer_value(j_value)) > 0) {
+			if (sprintf(s_int, "%" JSON_INTEGER_FORMAT,
+					json_integer_value(j_value)) > 0) {
 				/* set long value in the application header whose name is based on the key and the prefix */
-				oidc_util_set_app_header(r, s_key,
-						s_int,
-						claim_prefix);
+				oidc_util_set_app_header(r, s_key, s_int, claim_prefix);
 			} else {
-				ap_log_rerror(APLOG_MARK, APLOG_WARNING, 0, r,
-						"oidc_util_set_app_headers: could not convert JSON number to string (> 255 characters?), skipping");
+				oidc_warn(r,
+						"could not convert JSON number to string (> 255 characters?), skipping");
 			}
 
 		} else if (json_is_real(j_value)) {
@@ -1169,12 +1188,19 @@ void oidc_util_set_app_headers(request_rec *r, const json_t *j_attrs, const char
 					apr_psprintf(r->pool, "%lf", json_real_value(j_value)),
 					claim_prefix);
 
+		} else if (json_is_object(j_value)) {
+
+			/* set json value in the application header whose name is based on the key and the prefix */
+			char *s_value = json_dumps(j_value, 0);
+			oidc_util_set_app_header(r, s_key, s_value, claim_prefix);
+			free(s_value);
+
 			/* check if it is a multi-value string */
 		} else if (json_is_array(j_value)) {
 
 			/* some logging about what we're going to do */
-			ap_log_rerror(APLOG_MARK, OIDC_DEBUG, 0, r,
-					"oidc_util_set_app_headers: parsing attribute array for key \"%s\" (#nr-of-elems: %zu)",
+			oidc_debug(r,
+					"parsing attribute array for key \"%s\" (#nr-of-elems: %zu)",
 					s_key, json_array_size(j_value));
 
 			/* string to hold the concatenated array string values */
@@ -1208,14 +1234,14 @@ void oidc_util_set_app_headers(request_rec *r, const json_t *j_attrs, const char
 								json_is_true(elem) ? "1" : "0");
 					} else {
 						s_concat = apr_psprintf(r->pool, "%s",
-								json_is_true(elem) ? "1" : "0");
+						json_is_true(elem) ? "1" : "0");
 					}
 
 				} else {
 
 					/* don't know how to handle a non-string array element */
-					ap_log_rerror(APLOG_MARK, APLOG_WARNING, 0, r,
-							"oidc_util_set_app_headers: unhandled in-array JSON object type [%d] for key \"%s\" when parsing claims array elements",
+					oidc_warn(r,
+							"unhandled in-array JSON object type [%d] for key \"%s\" when parsing claims array elements",
 							elem->type, s_key);
 				}
 			}
@@ -1226,12 +1252,12 @@ void oidc_util_set_app_headers(request_rec *r, const json_t *j_attrs, const char
 		} else {
 
 			/* no string and no array, so unclear how to handle this */
-			ap_log_rerror(APLOG_MARK, APLOG_WARNING, 0, r,
-					"oidc_util_set_app_headers: unhandled JSON object type [%d] for key \"%s\" when parsing claims",
+			oidc_warn(r,
+					"unhandled JSON object type [%d] for key \"%s\" when parsing claims",
 					j_value->type, s_key);
 		}
 
-		iter = json_object_iter_next((json_t *)j_attrs, iter);
+		iter = json_object_iter_next((json_t *) j_attrs, iter);
 	}
 }
 
